"""Summary for rdbms database."""

import re
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

from dbgpt._private.config import Config
from dbgpt.datasource import BaseConnector
from dbgpt.rag.summary.db_summary import DBSummary

if TYPE_CHECKING:
    from dbgpt.datasource.manages import ConnectorManager

CFG = Config()


_DEFAULT_SUMMARY_TEMPLATE = """\
table_name: {table_name}\r\n\
table_comment: {table_comment}\r\n\
index_keys: {index_keys}\r\n\
"""
_DEFAULT_SUMMARY_TEMPLATE_PATTEN = (
    r"table_name:\s*(?P<table_name>.*)\s*"
    r"table_comment:\s*(?P<table_comment>.*)\s*"
    r"index_keys:\s*(?P<index_keys>.*)\s*"
)
_DEFAULT_COLUMN_SEPARATOR = ",\r\n    "


def _parse_table_detail(table_desc_str: str) -> Dict[str, Any]:
    """Parse table detail string.

    Args:
        table_desc_str (str): table detail string

    Returns:
        Dict[str, Any]: A dictionary containing table_name, table_comment, and
            index_keys.
    """
    matched = re.match(_DEFAULT_SUMMARY_TEMPLATE_PATTEN, table_desc_str)
    if matched:
        return matched.groupdict()
    return {}


class RdbmsSummary(DBSummary):
    """Get rdbms db table summary template.

    Summary example:
        table_name(column1(column1 comment),column2(column2 comment),
        column3(column3 comment) and index keys, and table comment is {table_comment})
    """

    def __init__(
        self, name: str, type: str, manager: Optional["ConnectorManager"] = None
    ):
        """Create a new RdbmsSummary."""
        self.name = name
        self.type = type
        self.summary_template = "{table_name}({columns})"
        self.tables = {}
        # self.tables_info = []
        # self.vector_tables_info = []

        # TODO: Don't use the global variable.
        db_manager = manager or CFG.local_db_manager
        if not db_manager:
            raise ValueError("Local db manage is not initialized.")
        self.db = db_manager.get_connector(name)

        self.metadata = """user info :{users}, grant info:{grant}, charset:{charset},
        collation:{collation}""".format(
            users=self.db.get_users(),
            grant=self.db.get_grants(),
            charset=self.db.get_charset(),
            collation=self.db.get_collation(),
        )
        tables = self.db.get_table_names()
        self.table_info_summaries = [
            self.get_table_summary(table_name) for table_name in tables
        ]

    def get_table_summary(self, table_name):
        """Get table summary for table.

        example:
            table_name(column1(column1 comment),column2(column2 comment),
            column3(column3 comment) and index keys, and table comment: {table_comment})
        """
        return _parse_table_summary(self.db, self.summary_template, table_name)

    def table_summaries(self):
        """Get table summaries."""
        return self.table_info_summaries


def _parse_db_summary(
    conn: BaseConnector, summary_template: str = "{table_name}({columns})"
) -> List[str]:
    """Get db summary for database.

    Args:
        conn (BaseConnector): database connection
        summary_template (str): summary template
    """
    tables = conn.get_table_names()
    table_info_summaries = [
        _parse_table_summary(conn, summary_template, table_name)
        for table_name in tables
    ]
    return table_info_summaries


def _parse_db_summary_with_metadata(
    conn: BaseConnector,
    summary_template: str = _DEFAULT_SUMMARY_TEMPLATE,
    separator: str = "--table-field-separator--",
    column_separator: str = _DEFAULT_COLUMN_SEPARATOR,
    model_dimension: int = 512,
) -> List[Tuple[str, Dict[str, Any]]]:
    """Get db summary for database.

    Args:
        conn (BaseConnector): database connection
        summary_template (str): summary template
        separator(str, optional): separator used to separate table's
            basic info and fields. defaults to `-- table-field-separator--`
        model_dimension(int, optional): The threshold for splitting field string
    """
    tables = conn.get_table_names()
    table_info_summaries = [
        _parse_table_summary_with_metadata(
            conn,
            summary_template,
            separator,
            table_name,
            model_dimension,
            column_separator=column_separator,
        )
        for table_name in tables
    ]
    return table_info_summaries


def _split_columns_str(
    columns: List[str], model_dimension: int, column_separator: str = ",\r\n    "
):
    """Split columns str.

    Args:
    columns (List[str]): fields string
    model_dimension (int, optional): The threshold for splitting field string.
    """
    result = []
    current_string = ""
    current_length = 0

    for element_str in columns:
        element_length = len(element_str)

        # If adding the current element's length would exceed the threshold,
        # add the current string to results and reset
        if current_length + element_length > model_dimension:
            result.append(current_string.strip())  # Remove trailing spaces
            current_string = element_str
            current_length = element_length
        else:
            # If current string is empty, add element directly
            if current_string:
                current_string += column_separator + element_str
            else:
                current_string = element_str
            current_length += element_length + 1  # Add length of space

    # Handle the last string segment
    if current_string:
        result.append(current_string.strip())

    return result


def _parse_table_summary_with_metadata(
    conn: BaseConnector,
    summary_template: str,
    separator,
    table_name: str,
    model_dimension=512,
    column_separator: str = _DEFAULT_COLUMN_SEPARATOR,
    db_summary_version: str = "v1.0",
) -> Tuple[str, Dict[str, Any]]:
    """Get table summary for table.

    Args:
        conn (BaseConnector): database connection
        summary_template (str): summary template
        separator(str, optional): separator used to separate table's
            basic info and fields. defaults to `-- table-field-separator--`
        model_dimension(int, optional): The threshold for splitting field string

    Examples:
        metadata: {'table_name': 'asd', 'separated': 0/1}

        table_name: table1
        table_comment: comment
        index_keys: keys
        --table-field-separator--
        (column1,comment), (column2, comment), (column3, comment)
        (column4,comment), (column5, comment), (column6, comment)
    """
    columns = []
    metadata = {
        "table_name": table_name,
        "separated": 0,
        "db_summary_version": db_summary_version,
    }
    for column in conn.get_columns(table_name):
        col_name = column["name"]
        col_type = str(column["type"]) if "type" in column else None
        col_comment = column.get("comment")
        column_def = f'"{col_name}" {col_type.upper()}'
        if col_comment:
            column_def += f' COMMENT "{col_comment}"'
        columns.append(column_def)
    metadata.update({"field_num": len(columns)})
    separated_columns = _split_columns_str(
        columns, model_dimension=model_dimension, column_separator=column_separator
    )
    if len(separated_columns) > 1:
        metadata["separated"] = 1
    column_str = column_separator.join(separated_columns)
    # Obtain index information
    index_keys = []
    raw_indexes = conn.get_indexes(table_name)
    for index in raw_indexes:
        if isinstance(index, tuple):  # Process tuple type index information
            index_name, index_creation_command = index
            # Extract column names using re
            matched_columns = re.findall(r"\(([^)]+)\)", index_creation_command)
            if matched_columns:
                key_str = ", ".join(matched_columns)
                index_keys.append(f"{index_name}(`{key_str}`) ")
        else:
            key_str = ", ".join(index["column_names"])
            index_keys.append(f"{index['name']}(`{key_str}`) ")

    table_comment = ""

    try:
        comment = conn.get_table_comment(table_name)
        table_comment = comment.get("text")
    except Exception:
        pass

    index_key_str = ", ".join(index_keys)
    table_str = summary_template.format(
        table_name=table_name, table_comment=table_comment, index_keys=index_key_str
    )
    table_str += f"\n{separator}\n{column_str}"
    return table_str, metadata


def _parse_table_summary(
    conn: BaseConnector, summary_template: str, table_name: str
) -> str:
    """Get table summary for table.

    Args:
        conn (BaseConnector): database connection
        summary_template (str): summary template
        table_name (str): table name

    Examples:
        table_name(column1(column1 comment),column2(column2 comment),
        column3(column3 comment) and index keys, and table comment: {table_comment})
    """
    columns = []
    for column in conn.get_columns(table_name):
        if column.get("comment"):
            columns.append(f"{column['name']} ({column.get('comment')})")
        else:
            columns.append(f"{column['name']}")

    column_str = ", ".join(columns)
    # Obtain index information
    index_keys = []
    raw_indexes = conn.get_indexes(table_name)
    for index in raw_indexes:
        if isinstance(index, tuple):  # Process tuple type index information
            index_name, index_creation_command = index
            # Extract column names using re
            matched_columns = re.findall(r"\(([^)]+)\)", index_creation_command)
            if matched_columns:
                key_str = ", ".join(matched_columns)
                index_keys.append(f"{index_name}(`{key_str}`) ")
        else:
            key_str = ", ".join(index["column_names"])
            index_keys.append(f"{index['name']}(`{key_str}`) ")
    table_str = summary_template.format(table_name=table_name, columns=column_str)
    if len(index_keys) > 0:
        index_key_str = ", ".join(index_keys)
        table_str += f", and index keys: {index_key_str}"
    try:
        comment = conn.get_table_comment(table_name)
    except Exception:
        comment = dict(text=None)
    if comment.get("text"):
        table_str += f", and table comment: {comment.get('text')}"
    return table_str



"""Summary for rdbms database."""

from typing import TYPE_CHECKING, Dict, List, Optional

from dbgpt._private.config import Config
from dbgpt.datasource.base import BaseConnector
from dbgpt.rag.summary.db_summary import DBSummary
from dbgpt_ext.datasource.conn_tugraph import TuGraphConnector

if TYPE_CHECKING:
    from dbgpt.datasource.manages import ConnectorManager

CFG = Config()


class GdbmsSummary(DBSummary):
    """Get graph db table summary template."""

    def __init__(
        self, name: str, type: str, manager: Optional["ConnectorManager"] = None
    ):
        """Create a new RdbmsSummary."""
        self.name = name
        self.type = type
        self.summary_template = "{table_name}({columns})"
        # self.v_summary_template = "{table_name}({columns})"
        self.tables = {}
        # self.tables_info = []
        # self.vector_tables_info = []

        # TODO: Don't use the global variable.
        db_manager = manager or CFG.local_db_manager
        if not db_manager:
            raise ValueError("Local db manage is not initialized.")
        self.db = db_manager.get_connector(name)

        self.metadata = """user info :{users}, grant info:{grant}, charset:{charset},
        collation:{collation}""".format(
            users=self.db.get_users(),
            grant=self.db.get_grants(),
            charset=self.db.get_charset(),
            collation=self.db.get_collation(),
        )
        tables = self.db.get_table_names()
        self.table_info_summaries = {
            "vertex_tables": [
                self.get_table_summary(table_name.split("_")[0], "vertex")
                for table_name in tables
                if table_name.endswith("_vertex")
            ],
            "edge_tables": [
                self.get_table_summary(table_name.split("_")[0], "edge")
                for table_name in tables
                if table_name.endswith("_edge")
            ],
        }

    def get_table_summary(self, table_name, table_type):
        """Get table summary for table.

        example:
            table_name(column1(column1 comment),column2(column2 comment),
            column3(column3 comment) and index keys, and table comment: {table_comment})
        """
        return _parse_table_summary(
            self.db, self.summary_template, table_name, table_type
        )

    def table_summaries(self):
        """Get table summaries."""
        return self.table_info_summaries


def _parse_db_summary(
    conn: BaseConnector, summary_template: str = "{table_name}({columns})"
) -> List[str]:
    """Get db summary for database."""
    table_info_summaries = None
    if isinstance(conn, TuGraphConnector):
        table_names = conn.get_table_names()
        v_tables = [
            table_name.split("_")[0]
            for table_name in table_names
            if table_name.endswith("_vertex")
        ]
        e_tables = [
            table_name.split("_")[0]
            for table_name in table_names
            if table_name.endswith("_edge")
        ]
        table_info_summaries = [
            _parse_table_summary(conn, summary_template, table_name, "vertex")
            for table_name in v_tables
        ] + [
            _parse_table_summary(conn, summary_template, table_name, "edge")
            for table_name in e_tables
        ]
    else:
        table_info_summaries = []

    return table_info_summaries


def _format_column(column: Dict) -> str:
    """Format a single column's summary."""
    comment = column.get("comment", "")
    if column.get("is_in_primary_key"):
        comment += " Primary Key" if comment else "Primary Key"
    return f"{column['name']} ({comment})" if comment else column["name"]


def _format_indexes(indexes: List[Dict]) -> str:
    """Format index keys for table summary."""
    return ", ".join(
        f"{index['name']}(`{', '.join(index['column_names'])}`)" for index in indexes
    )


def _parse_table_summary(
    conn: TuGraphConnector, summary_template: str, table_name: str, table_type: str
) -> str:
    """Enhanced table summary function."""
    columns = [
        _format_column(column) for column in conn.get_columns(table_name, table_type)
    ]
    column_str = ", ".join(columns)

    indexes = conn.get_indexes(table_name, table_type)
    index_str = _format_indexes(indexes) if indexes else ""

    table_str = summary_template.format(table_name=table_name, columns=column_str)
    if index_str:
        table_str += f", and index keys: {index_str}"
    try:
        comment = conn.get_table_comment(table_name)
    except Exception:
        comment = dict(text=None)
    if comment.get("text"):
        table_str += (
            f", and table comment: {comment.get('text')}, this is a {table_type} table"
        )
    else:
        table_str += f", and table comment: this is a {table_type} table"
    return table_str



"""
Run unit test with command: pytest dbgpt/datasource/rdbms/tests/test_conn_duckdb.py
"""

import tempfile

import pytest

from dbgpt_ext.datasource.rdbms.conn_duckdb import DuckDbConnector


@pytest.fixture
def db():
    temp_db_file = tempfile.NamedTemporaryFile(delete=False)
    temp_db_file.close()
    conn = DuckDbConnector.from_file_path(temp_db_file.name + "duckdb.db")
    yield conn


def test_get_users(db):
    assert db.get_users() == []


def test_get_table_names(db):
    assert list(db.get_table_names()) == []


def test_get_charset(db):
    assert db.get_charset() == "UTF-8"


def test_get_table_comments(db):
    assert db.get_table_comments("test") == []


def test_table_simple_info(db):
    assert db.table_simple_info() == []


def test_execute(db):
    assert list(db.run("SELECT 42")[0]) == ["42"]



"""Aliyun OSS storage backend."""

import hashlib
import io
import logging
import os
import random
import time
from typing import BinaryIO, Callable, Dict, Optional, Union

import oss2
from oss2.credentials import EnvironmentVariableCredentialsProvider

from dbgpt.core.interface.file import FileMetadata, StorageBackend

logger = logging.getLogger(__name__)


def does_bucket_exist(bucket):
    try:
        bucket.get_bucket_info()
    except oss2.exceptions.NoSuchBucket:
        return False
    except:
        raise
    return True


class AliyunOSSStorage(StorageBackend):
    """Aliyun OSS storage backend implementation."""

    storage_type: str = "oss"

    def __init__(
        self,
        endpoint: str,
        region: str,
        access_key_id: Optional[str] = None,
        access_key_secret: Optional[str] = None,
        save_chunk_size: int = 1024 * 1024,
        use_environment_credentials: bool = False,
        fixed_bucket: Optional[str] = None,
        bucket_prefix: str = "dbgpt-fs-",
        bucket_mapper: Optional[Callable[[str], str]] = None,
        auto_create_bucket: bool = True,
        default_public_url_expire: int = 3600,
    ):
        """Initialize the Aliyun OSS storage backend.

        Args:
            endpoint (str): OSS endpoint, e.g., "https://oss-cn-hangzhou.aliyuncs.com"
            region (str): OSS region, e.g., "cn-hangzhou"
            access_key_id (Optional[str], optional): Aliyun Access Key ID. Defaults to
                None.
            access_key_secret (Optional[str], optional): Aliyun Access Key Secret.
                Defaults to None.
            save_chunk_size (int, optional): Chunk size for saving files. Defaults to
                1024*1024 (1MB).
            use_environment_credentials (bool, optional): Whether to use credentials
                from environment variables. Defaults to False.
            fixed_bucket (Optional[str], optional): A fixed OSS bucket to use for all
                operations. If provided, all logical buckets will be mapped to this
                single bucket. Defaults to None.
            bucket_prefix (str, optional): Prefix for dynamically created buckets.
                Defaults to "dbgpt-fs-".
            bucket_mapper (Optional[Callable[[str], str]], optional): Custom function
                to map logical bucket names to actual OSS bucket names. Defaults to
                None.
            auto_create_bucket (bool, optional): Whether to automatically create
                buckets that don't exist. Defaults to True.
        """
        self.endpoint = endpoint
        self.region = region
        self._save_chunk_size = save_chunk_size
        self.fixed_bucket = fixed_bucket
        self.bucket_prefix = bucket_prefix
        self.custom_bucket_mapper = bucket_mapper
        self.auto_create_bucket = auto_create_bucket
        self.default_public_url_expire = default_public_url_expire

        # Initialize OSS authentication
        if use_environment_credentials:
            # Check required environment variables
            required_env_vars = ["OSS_ACCESS_KEY_ID", "OSS_ACCESS_KEY_SECRET"]
            for var in required_env_vars:
                if var not in os.environ:
                    raise ValueError(f"Environment variable {var} is not set.")
            self.auth = oss2.ProviderAuthV4(EnvironmentVariableCredentialsProvider())
        else:
            if not access_key_id or not access_key_secret:
                raise ValueError(
                    "Access key ID and secret are required when not using environment "
                    "credentials"
                )
            # Use provided credentials
            self.auth = oss2.Auth(access_key_id, access_key_secret)

        # Store buckets dict to avoid recreating bucket objects
        self._buckets: Dict[str, oss2.Bucket] = {}

        # Create fixed bucket if specified
        if self.fixed_bucket and self.auto_create_bucket:
            self._ensure_bucket_exists(self.fixed_bucket)

    @property
    def save_chunk_size(self) -> int:
        """Get the save chunk size."""
        return self._save_chunk_size

    def _map_bucket_name(self, logical_bucket: str) -> str:
        """Map logical bucket name to actual OSS bucket name.

        Args:
            logical_bucket (str): Logical bucket name used by the application

        Returns:
            str: Actual OSS bucket name to use
        """
        # 1. If using a fixed bucket, always return that
        if self.fixed_bucket:
            return self.fixed_bucket

        # 2. If a custom mapper is provided, use that
        if self.custom_bucket_mapper:
            return self.custom_bucket_mapper(logical_bucket)

        # 3. Otherwise, use a hash-based approach to generate a unique but
        # deterministic name
        # This avoids bucket name conflicts while maintaining consistency
        bucket_hash = hashlib.md5(logical_bucket.encode()).hexdigest()[:8]
        return f"{self.bucket_prefix}{bucket_hash}-{logical_bucket}"

    def _generate_dynamic_bucket_name(self) -> str:
        """Generate a unique bucket name for dynamic creation.

        Returns:
            str: A unique bucket name
        """
        # Using timestamp + random number to ensure uniqueness
        timestamp = int(time.time())
        random_number = random.randint(0, 9999)
        return f"{self.bucket_prefix}{timestamp}-{random_number}"

    def _ensure_bucket_exists(self, bucket_name: str) -> bool:
        """Ensure the bucket exists, create it if needed and if auto_create_bucket is
        True.

        Args:
            bucket_name (str): Bucket name

        Returns:
            bool: True if the bucket exists or was created, False otherwise
        """
        bucket = oss2.Bucket(self.auth, self.endpoint, bucket_name, region=self.region)

        try:
            if does_bucket_exist(bucket):
                return True

            if not self.auto_create_bucket:
                logger.warning(
                    f"Bucket {bucket_name} does not exist and auto_create_bucket is "
                    f"False"
                )
                return False

            logger.info(f"Creating bucket {bucket_name}")
            bucket.create_bucket(oss2.models.BUCKET_ACL_PRIVATE)
            return True
        except oss2.exceptions.ServerError as e:
            # Handle the case where bucket name is already taken by someone else
            if e.status == 409 and "BucketAlreadyExists" in str(e):
                logger.warning(
                    f"Bucket name {bucket_name} already exists and is owned by "
                    "someone else"
                )
                return False
            raise
        except oss2.exceptions.OssError as e:
            logger.error(f"Failed to create or check bucket {bucket_name}: {e}")
            raise

    def _get_bucket(self, logical_bucket: str) -> Union[oss2.Bucket, None]:
        """Get or create an OSS bucket object for the given logical bucket.

        Args:
            logical_bucket (str): Logical bucket name

        Returns:
            Union[oss2.Bucket, None]: Bucket object or None if bucket creation failed
        """
        # Get the actual OSS bucket name
        actual_bucket_name = self._map_bucket_name(logical_bucket)

        # Check if we've already cached this bucket
        if actual_bucket_name in self._buckets:
            return self._buckets[actual_bucket_name]

        # Try to ensure the mapped bucket exists
        if self._ensure_bucket_exists(actual_bucket_name):
            # Cache and return the bucket
            self._buckets[actual_bucket_name] = oss2.Bucket(
                self.auth, self.endpoint, actual_bucket_name, region=self.region
            )
            return self._buckets[actual_bucket_name]

        # If we get here, the bucket doesn't exist and couldn't be created
        # Try to create a dynamic bucket if we're not using a fixed bucket
        if not self.fixed_bucket and self.auto_create_bucket:
            # Generate a new unique bucket name
            dynamic_bucket = self._generate_dynamic_bucket_name()
            logger.info(
                f"Attempting to create dynamic bucket {dynamic_bucket} for logical "
                f"bucket {logical_bucket}"
            )

            if self._ensure_bucket_exists(dynamic_bucket):
                self._buckets[actual_bucket_name] = oss2.Bucket(
                    self.auth, self.endpoint, dynamic_bucket, region=self.region
                )
                return self._buckets[actual_bucket_name]

        # If all attempts failed
        raise ValueError(
            f"Failed to get or create bucket for logical bucket {logical_bucket}"
        )

    def save(
        self,
        bucket: str,
        file_id: str,
        file_data: BinaryIO,
        public_url: bool = False,
        public_url_expire: Optional[int] = None,
    ) -> str:
        """Save the file data to Aliyun OSS.

        Args:
            bucket (str): The logical bucket name
            file_id (str): The file ID
            file_data (BinaryIO): The file data
            public_url (bool, optional): Whether to generate a public URL. Defaults to
                False.
            public_url_expire (Optional[int], optional): Expiration time for the public
                URL in seconds. Defaults to None.

        Returns:
            str: The storage path (OSS URI)
        """
        # Get the actual OSS bucket
        oss_bucket = self._get_bucket(bucket)

        # Generate OSS object name based on whether we're using fixed bucket
        object_name = file_id
        if self.fixed_bucket:
            # When using a fixed bucket, we need to prefix with logical bucket name to
            # avoid conflicts
            object_name = f"{bucket}/{file_id}"

        # For large files, use multipart upload
        file_size = self._get_file_size(file_data)

        if file_size > 10 * self.save_chunk_size:  # If file is larger than 10MB
            logger.info(
                f"Using multipart upload for large file: {object_name} "
                f"(size: {file_size})"
            )
            self._multipart_upload(oss_bucket, object_name, file_data)
        else:
            logger.info(f"Uploading file using simple upload: {object_name}")
            try:
                oss_bucket.put_object(object_name, file_data)
            except oss2.exceptions.OssError as e:
                logger.error(
                    f"Failed to upload file {object_name} to bucket "
                    f"{oss_bucket.bucket_name}: {e}"
                )
                raise

        # Store the OSS bucket name and object path for future reference
        actual_bucket_name = oss_bucket.bucket_name

        # Format: oss://{actual_bucket_name}/{object_name}
        # We store both the actual bucket name and the object path in the URI
        # But we'll also keep the logical bucket in the external URI format
        storage_path = f"oss://{bucket}/{file_id}?actual_bucket={actual_bucket_name}&object_name={object_name}"  # noqa

        # Generate a public URL if requested
        if public_url:
            # Use provided expiration time or default
            expire_seconds = public_url_expire or self.default_public_url_expire

            # Generate a signed URL for public access
            try:
                url = oss_bucket.sign_url(
                    "GET", object_name, expire_seconds, slash_safe=True
                )
                logger.info(
                    f"Generated public URL for {object_name} with expiration "
                    f"{expire_seconds} seconds"
                )
                return url
            except oss2.exceptions.OssError as e:
                logger.error(f"Failed to generate public URL for {object_name}: {e}")
                # Fall back to returning the storage path
                return storage_path

        return storage_path

    def _get_file_size(self, file_data: BinaryIO) -> int:
        """Get file size without consuming the file object.

        Args:
            file_data (BinaryIO): The file data

        Returns:
            int: The file size in bytes
        """
        current_pos = file_data.tell()
        file_data.seek(0, io.SEEK_END)
        size = file_data.tell()
        file_data.seek(current_pos)  # Reset the file pointer
        return size

    def _multipart_upload(
        self, oss_bucket: oss2.Bucket, file_id: str, file_data: BinaryIO
    ) -> None:
        """Handle multipart upload for large files.

        Args:
            oss_bucket (oss2.Bucket): OSS bucket object
            file_id (str): The file ID
            file_data (BinaryIO): The file data
        """
        # Initialize multipart upload
        upload_id = oss_bucket.init_multipart_upload(file_id).upload_id

        # Upload parts
        part_number = 1
        parts = []

        while True:
            chunk = file_data.read(self.save_chunk_size)
            if not chunk:
                break

            # Upload part
            etag = oss_bucket.upload_part(file_id, upload_id, part_number, chunk).etag
            parts.append(oss2.models.PartInfo(part_number, etag))
            part_number += 1

        # Complete multipart upload
        oss_bucket.complete_multipart_upload(file_id, upload_id, parts)

    def _parse_storage_path(self, storage_path: str) -> Dict[str, str]:
        """Parse the OSS storage path to extract actual bucket and object name.

        Args:
            storage_path (str): The storage path URI

        Returns:
            Dict[str, str]: A dictionary with actual_bucket and object_name keys
        """
        if not storage_path.startswith("oss://"):
            raise ValueError(f"Invalid storage path for Aliyun OSS: {storage_path}")

        # Example URI:
        # oss://logical_bucket/file_id?actual_bucket=oss_bucket&object_name=logical_bucket/file_id # noqa

        # Try to parse the URL parameters
        from urllib.parse import parse_qs, urlparse

        parsed_url = urlparse(storage_path)
        params = parse_qs(parsed_url.query)

        # Extract the parameters
        actual_bucket = params.get("actual_bucket", [None])[0]
        object_name = params.get("object_name", [None])[0]

        # Extract the logical bucket and file_id from the path
        path_parts = parsed_url.path.strip("/").split("/", 1)
        logical_bucket = path_parts[0] if path_parts else None
        logical_file_id = path_parts[1] if len(path_parts) > 1 else None

        # If parameters aren't in the URL (backward compatibility or simplified URL),
        # derive them from the logical values
        if not actual_bucket:
            # Try to use the bucket mapper to get the actual bucket
            actual_bucket = (
                self._map_bucket_name(logical_bucket) if logical_bucket else None
            )

        if not object_name:
            # If using fixed bucket, the object name includes the logical bucket
            # as prefix
            if self.fixed_bucket:
                object_name = (
                    f"{logical_bucket}/{logical_file_id}"
                    if logical_bucket and logical_file_id
                    else None
                )
            else:
                object_name = logical_file_id

        return {
            "logical_bucket": logical_bucket,
            "logical_file_id": logical_file_id,
            "actual_bucket": actual_bucket,
            "object_name": object_name,
        }

    def load(self, fm: FileMetadata) -> BinaryIO:
        """Load the file data from Aliyun OSS.

        Args:
            fm (FileMetadata): The file metadata

        Returns:
            BinaryIO: The file data as a binary IO object
        """
        # Parse the storage path
        path_info = self._parse_storage_path(fm.storage_path)

        # Get actual bucket and object name
        actual_bucket_name = path_info["actual_bucket"]
        object_name = path_info["object_name"]
        logical_bucket = path_info["logical_bucket"]

        # If we couldn't determine the actual bucket from the URI, try with the
        # logical bucket
        if not actual_bucket_name and logical_bucket:
            actual_bucket_name = self._map_bucket_name(logical_bucket)

        # Use the file_id as object name if object_name is still None
        if not object_name:
            object_name = fm.file_id
            # If using fixed bucket, prefix with logical bucket
            if self.fixed_bucket and logical_bucket:
                object_name = f"{logical_bucket}/{fm.file_id}"

        # Get the bucket object
        try:
            oss_bucket = oss2.Bucket(
                self.auth, self.endpoint, actual_bucket_name, region=self.region
            )

            # Get object as stream
            object_stream = oss_bucket.get_object(object_name)

            # Convert to BytesIO for compatibility
            content = io.BytesIO(object_stream.read())
            content.seek(0)
            return content
        except oss2.exceptions.NoSuchKey as e:
            logger.error(
                f"File {object_name} not found in bucket {actual_bucket_name}: {e}"
            )
            raise FileNotFoundError(
                f"File {object_name} not found in bucket {actual_bucket_name}"
            )
        except oss2.exceptions.OssError as e:
            logger.error(
                f"Failed to download file {object_name} from bucket "
                f"{actual_bucket_name}: {e}"
            )
            raise

    def delete(self, fm: FileMetadata) -> bool:
        """Delete the file data from Aliyun OSS.

        Args:
            fm (FileMetadata): The file metadata

        Returns:
            bool: True if the file was deleted, False otherwise
        """
        # Parse the storage path
        path_info = self._parse_storage_path(fm.storage_path)

        # Get actual bucket and object name
        actual_bucket_name = path_info["actual_bucket"]
        object_name = path_info["object_name"]
        logical_bucket = path_info["logical_bucket"]

        # If we couldn't determine the actual bucket from the URI, try with the
        # logical bucket
        if not actual_bucket_name and logical_bucket:
            actual_bucket_name = self._map_bucket_name(logical_bucket)

        # Use the file_id as object name if object_name is still None
        if not object_name:
            object_name = fm.file_id
            # If using fixed bucket, prefix with logical bucket
            if self.fixed_bucket and logical_bucket:
                object_name = f"{logical_bucket}/{fm.file_id}"

        try:
            # Get the bucket object
            oss_bucket = oss2.Bucket(
                self.auth, self.endpoint, actual_bucket_name, region=self.region
            )

            # Check if the object exists
            if not oss_bucket.object_exists(object_name):
                logger.warning(
                    f"File {object_name} does not exist in bucket {actual_bucket_name}"
                )
                return False

            # Delete the object
            oss_bucket.delete_object(object_name)
            logger.info(f"File {object_name} deleted from bucket {actual_bucket_name}")
            return True
        except oss2.exceptions.OssError as e:
            logger.error(
                f"Failed to delete file {object_name} from bucket {actual_bucket_name}:"
                f" {e}"
            )
            return False

    def get_public_url(
        self, fm: FileMetadata, expire: Optional[int] = None
    ) -> Optional[str]:
        """Generate a public URL for an existing file.

        Args:
            fm (FileMetadata): The file metadata
            expire (Optional[int], optional): Expiration time in seconds. Defaults to
                class default.

        Returns:
            str: The generated public URL
        """
        # Parse the storage path
        path_info = self._parse_storage_path(fm.storage_path)

        # Get actual bucket and object name
        actual_bucket_name = path_info["actual_bucket"]
        object_name = path_info["object_name"]
        logical_bucket = path_info["logical_bucket"]

        # If we couldn't determine the actual bucket from the URI, try with the logical
        # bucket
        if not actual_bucket_name and logical_bucket:
            actual_bucket_name = self._map_bucket_name(logical_bucket)

        # Use the file_id as object name if object_name is still None
        if not object_name:
            object_name = fm.file_id
            # If using fixed bucket, prefix with logical bucket
            if self.fixed_bucket and logical_bucket:
                object_name = f"{logical_bucket}/{fm.file_id}"

        # Get the bucket object
        try:
            oss_bucket = oss2.Bucket(
                self.auth, self.endpoint, actual_bucket_name, region=self.region
            )

            # Use provided expiration time or default
            expire_seconds = expire or self.default_public_url_expire

            # Generate signed URL
            url = oss_bucket.sign_url(
                "GET", object_name, expire_seconds, slash_safe=True
            )
            logger.info(
                f"Generated public URL for {object_name} with expiration "
                f"{expire_seconds} seconds"
            )
            return url
        except oss2.exceptions.OssError as e:
            logger.error(f"Failed to generate public URL for {fm.file_id}: {e}")
            raise



"""TextEmbedder class."""

import asyncio
import logging
from typing import List

from dbgpt.rag.transformer.base import EmbedderBase

logger = logging.getLogger(__name__)


class TextEmbedder(EmbedderBase):
    """TextEmbedder class."""

    async def batch_embed(
        self,
        inputs: List[str],
        batch_size: int = 1,
    ) -> List[List[float]]:
        """Embed texts from graphs in batches."""
        vectors: List[List[float]] = []
        n_texts = len(inputs)

        # Batch embedding
        for batch_idx in range(0, n_texts, batch_size):
            start_idx = batch_idx
            end_idx = min(start_idx + batch_size, n_texts)
            batch_texts = inputs[start_idx:end_idx]

            # Create tasks
            embedding_tasks = [(self.embed(text)) for text in batch_texts]

            # Process embedding in parallel
            batch_results = await asyncio.gather(
                *(task for task in embedding_tasks), return_exceptions=False
            )

            # Place results in the correct positions
            for idx, vector in enumerate(batch_results):
                if isinstance(vector, Exception):
                    raise RuntimeError(f"Failed to embed text{idx}")
                vectors.append(vector)

        return vectors

    def truncate(self):
        """Do nothing by default."""

    def drop(self):
        """Do nothing by default."""



"""OceanBase vector store."""

import json
import logging
import math
import os
import uuid
from dataclasses import dataclass, field
from typing import Any, List, Optional, Tuple

import numpy as np
from sqlalchemy import JSON, Column, String, Table, func, text
from sqlalchemy.dialects.mysql import LONGTEXT

from dbgpt.core import Chunk, Embeddings
from dbgpt.core.awel.flow import Parameter, ResourceCategory, register_resource
from dbgpt.storage.vector_store.base import (
    _COMMON_PARAMETERS,
    _VECTOR_STORE_COMMON_PARAMETERS,
    VectorStoreBase,
    VectorStoreConfig,
)
from dbgpt.storage.vector_store.filters import FilterOperator, MetadataFilters
from dbgpt.util.i18n_utils import _

logger = logging.getLogger(__name__)

DEFAULT_OCEANBASE_BATCH_SIZE = 100
DEFAULT_OCEANBASE_VECTOR_TABLE_NAME = "dbgpt_vector"
DEFAULT_OCEANBASE_HNSW_BUILD_PARAM = {"M": 16, "efConstruction": 256}
DEFAULT_OCEANBASE_HNSW_SEARCH_PARAM = {"efSearch": 64}
OCEANBASE_SUPPORTED_VECTOR_INDEX_TYPE = "HNSW"
DEFAULT_OCEANBASE_VECTOR_METRIC_TYPE = "l2"

DEFAULT_OCEANBASE_PFIELD = "id"
DEFAULT_OCEANBASE_DOCID_FIELD = "doc_id"
DEFAULT_OCEANBASE_VEC_FIELD = "embedding"
DEFAULT_OCEANBASE_DOC_FIELD = "document"
DEFAULT_OCEANBASE_METADATA_FIELD = "metadata"

DEFAULT_OCEANBASE_VEC_INDEX_NAME = "vidx"


def _parse_filter_value(filter_value: Any, is_text_match: bool = False):
    if filter_value is None:
        return filter_value

    if is_text_match:
        return f"'{filter_value!s}%'"

    if isinstance(filter_value, str):
        return f"'{filter_value!s}'"

    if isinstance(filter_value, list):
        if all(isinstance(item, str) for item in filter_value):
            return "(" + ",".join([f"'{str(v)}'" for v in filter_value]) + ")"
        return "(" + ",".join([str(v) for v in filter_value]) + ")"

    return str(filter_value)


def _euclidean_similarity(distance: float) -> float:
    return 1.0 - distance / math.sqrt(2)


def _neg_inner_product_similarity(distance: float) -> float:
    return -distance


def _normalize(vector: List[float]) -> List[float]:
    arr = np.array(vector)
    norm = np.linalg.norm(arr)
    arr = arr / norm
    return arr.tolist()


@register_resource(
    _("OceanBase Config"),
    "oceanbase_vector_config",
    category=ResourceCategory.VECTOR_STORE,
    parameters=[
        *_COMMON_PARAMETERS,
        Parameter.build_from(
            _("OceanBase Host"),
            "ob_host",
            str,
            description=_("oceanbase host"),
            optional=True,
            default=None,
        ),
        Parameter.build_from(
            _("OceanBase Port"),
            "ob_port",
            int,
            description=_("oceanbase port"),
            optional=True,
            default=None,
        ),
        Parameter.build_from(
            _("OceanBase User"),
            "ob_user",
            str,
            description=_("user to login"),
            optional=True,
            default=None,
        ),
        Parameter.build_from(
            _("OceanBase Password"),
            "ob_password",
            str,
            description=_("password to login"),
            optional=True,
            default=None,
        ),
        Parameter.build_from(
            _("OceanBase Database"),
            "ob_database",
            str,
            description=_("database for vector tables"),
            optional=True,
            default=None,
        ),
    ],
    description="OceanBase vector store config.",
)
@dataclass
class OceanBaseConfig(VectorStoreConfig):
    """OceanBase vector store config."""

    __type__ = "oceanbase"

    ob_host: Optional[str] = field(
        default=None,
        metadata={
            "help": "The host of oceanbase, if not set, will use the default host."
        },
    )
    ob_port: Optional[int] = field(
        default=None,
        metadata={
            "help": "The port of oceanbase, if not set, will use the default port."
        },
    )
    ob_user: Optional[str] = field(
        default=None,
        metadata={
            "help": "The user of oceanbase, if not set, will use the default user."
        },
    )
    ob_password: Optional[str] = field(
        default=None,
        metadata={
            "help": "The password of oceanbase, if not set, "
            "will use the default password"
        },
    )
    ob_database: Optional[str] = field(
        default=None,
        metadata={
            "help": "The database for vector tables, if not set, "
            "will use the default database."
        },
    )

    def create_store(self, **kwargs) -> "OceanBaseStore":
        """Create OceanBase store."""
        return OceanBaseStore(vector_store_config=self, **kwargs)


@register_resource(
    _("OceanBase Vector Store"),
    "ob_vector_store",
    category=ResourceCategory.VECTOR_STORE,
    description=_("OceanBase vector store."),
    parameters=[
        Parameter.build_from(
            _("OceanBase Config"),
            "vector_store_config",
            OceanBaseConfig,
            description=_("the ob config of vector store."),
            optional=True,
            default=None,
        ),
        *_VECTOR_STORE_COMMON_PARAMETERS,
    ],
)
class OceanBaseStore(VectorStoreBase):
    """OceanBase vector store."""

    def __init__(
        self,
        vector_store_config: OceanBaseConfig,
        name: Optional[str],
        embedding_fn: Optional[Embeddings] = None,
        max_chunks_once_load: Optional[int] = None,
        max_threads: Optional[int] = None,
    ) -> None:
        """Create a OceanBaseStore instance."""
        try:
            from pyobvector import ObVecClient  # type: ignore
        except ImportError:
            raise ImportError(
                "Could not import pyobvector package. "
                "Please install it with `pip install pyobvector`."
            )

        if embedding_fn is None:
            raise ValueError("embedding_fn is required for OceanBaseStore")

        super().__init__(
            max_chunks_once_load=max_chunks_once_load, max_threads=max_threads
        )

        self._vector_store_config = vector_store_config
        self.embedding_function = embedding_fn
        self.table_name = name

        vector_store_config_map = vector_store_config.to_dict()
        OB_HOST = str(
            vector_store_config_map.get("ob_host") or os.getenv("OB_HOST", "127.0.0.1")
        )
        OB_PORT = int(
            vector_store_config_map.get("ob_port") or int(os.getenv("OB_PORT", "2881"))
        )
        OB_USER = str(
            vector_store_config_map.get("ob_user") or os.getenv("OB_USER", "root@test")
        )
        OB_PASSWORD = str(
            vector_store_config_map.get("ob_password") or os.getenv("OB_PASSWORD", "")
        )
        OB_DATABASE = str(
            vector_store_config_map.get("ob_database")
            or os.getenv("OB_DATABASE", "test")
        )

        self.normalize = bool(os.getenv("OB_ENABLE_NORMALIZE_VECTOR", ""))
        self.vidx_metric_type = DEFAULT_OCEANBASE_VECTOR_METRIC_TYPE
        self.vidx_algo_params = DEFAULT_OCEANBASE_HNSW_BUILD_PARAM
        self.primary_field = DEFAULT_OCEANBASE_PFIELD
        self.vector_field = DEFAULT_OCEANBASE_VEC_FIELD
        self.text_field = DEFAULT_OCEANBASE_DOC_FIELD
        self.metadata_field = DEFAULT_OCEANBASE_METADATA_FIELD
        self.vidx_name = DEFAULT_OCEANBASE_VEC_INDEX_NAME
        self.hnsw_ef_search = -1

        self.vector_store_client = ObVecClient(
            uri=OB_HOST + ":" + str(OB_PORT),
            user=OB_USER,
            password=OB_PASSWORD,
            db_name=OB_DATABASE,
            echo=True,
        )

    def get_config(self) -> OceanBaseConfig:
        """Get the vector store config."""
        return self._vector_store_config

    def vector_name_exists(self) -> bool:
        """Whether vector name exists."""
        return self.vector_store_client.check_table_exists(table_name=self.table_name)

    def _load_table(self) -> None:
        table = Table(
            self.table_name,
            self.vector_store_client.metadata_obj,
            autoload_with=self.vector_store_client.engine,
        )
        column_names = [column.name for column in table.columns]
        assert len(column_names) == 4

        self.primary_field = column_names[0]
        self.vector_field = column_names[1]
        self.text_field = column_names[2]
        self.metadata_field = column_names[3]

    def _create_table_with_index(self, embeddings: list) -> None:
        try:
            from pyobvector import VECTOR
        except ImportError:
            raise ImportError(
                "Could not import pyobvector package. "
                "Please install it with `pip install pyobvector`."
            )

        if self.vector_store_client.check_table_exists(self.table_name):
            self._load_table()
            return

        dim = len(embeddings[0])
        cols = [
            Column(
                self.primary_field, String(4096), primary_key=True, autoincrement=False
            ),
            Column(self.vector_field, VECTOR(dim)),
            Column(self.text_field, LONGTEXT),
            Column(self.metadata_field, JSON),
        ]

        vidx_params = self.vector_store_client.prepare_index_params()
        vidx_params.add_index(
            field_name=self.vector_field,
            index_type=OCEANBASE_SUPPORTED_VECTOR_INDEX_TYPE,
            index_name=self.vidx_name,
            metric_type=self.vidx_metric_type,
            params=self.vidx_algo_params,
        )

        self.vector_store_client.create_table_with_index_params(
            table_name=self.table_name,
            columns=cols,
            indexes=None,
            vidxs=vidx_params,
        )

    def create_collection(self, collection_name: str, **kwargs) -> Any:
        """Create the collection."""
        embeddings = self.embedding_function.embed_documents([collection_name])
        return self._create_table_with_index(embeddings)

    def load_document(self, chunks: List[Chunk]) -> List[str]:
        """Load document in vector database."""
        batch_size = 100
        texts = [d.content for d in chunks]
        metadatas = [d.metadata for d in chunks]
        embeddings = self.embedding_function.embed_documents(texts)

        self._create_table_with_index(embeddings)

        ids = [str(uuid.uuid4()) for _ in texts]
        pks: list[str] = []
        for i in range(0, len(embeddings), batch_size):
            data = [
                {
                    self.primary_field: id,
                    self.vector_field: (
                        embedding if not self.normalize else _normalize(embedding)
                    ),
                    self.text_field: text,
                    self.metadata_field: metadata,
                }
                for id, embedding, text, metadata in zip(
                    ids[i : i + batch_size],
                    embeddings[i : i + batch_size],
                    texts[i : i + batch_size],
                    metadatas[i : i + batch_size],
                )
            ]
            self.vector_store_client.insert(
                table_name=self.table_name,
                data=data,
            )
            pks.extend(ids[i : i + batch_size])
        return pks

    def _parse_metric_type_str_to_dist_func(self) -> Any:
        if self.vidx_metric_type == "l2":
            return func.l2_distance
        if self.vidx_metric_type == "cosine":
            return func.cosine_distance
        if self.vidx_metric_type == "inner_product":
            return func.negative_inner_product
        raise ValueError(f"Invalid vector index metric type: {self.vidx_metric_type}")

    def similar_search(
        self,
        text: str,
        topk: int,
        filters: Optional[MetadataFilters] = None,
        param: Optional[dict] = None,
    ) -> List[Chunk]:
        """Perform a search on a query string and return results."""
        query_vector = self.embedding_function.embed_query(text)
        return self._similarity_search_by_vector(
            embedding=query_vector, k=topk, param=param, filters=filters
        )

    def similar_search_with_scores(
        self,
        text: str,
        topk: int,
        score_threshold: float,
        filters: Optional[MetadataFilters] = None,
        param: Optional[dict] = None,
    ) -> List[Chunk]:
        """Perform a search on a query string and return results with score."""
        query_vector = self.embedding_function.embed_query(text)
        docs_with_id_and_scores = self._similarity_search_with_score_by_vector(
            embedding=query_vector, k=topk, param=param, filters=filters
        )
        return [
            Chunk(
                metadata=doc.metadata,
                content=doc.content,
                score=score,
                chunk_id=str(id),
            )
            for doc, id, score in docs_with_id_and_scores
            if score >= score_threshold
        ]

    def _similarity_search_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        param: Optional[dict] = None,
        filters: Optional[MetadataFilters] = None,
    ) -> List[Chunk]:
        if filters is not None:
            filter = self._convert_metadata_filters(filters)
        else:
            filter = None

        search_param = (
            param if param is not None else DEFAULT_OCEANBASE_HNSW_SEARCH_PARAM
        )
        ef_search = search_param.get(
            "efSearch", DEFAULT_OCEANBASE_HNSW_SEARCH_PARAM["efSearch"]
        )
        if ef_search != self.hnsw_ef_search:
            self.vector_store_client.set_ob_hnsw_ef_search(ef_search)
            self.hnsw_ef_search = ef_search

        res = self.vector_store_client.ann_search(
            table_name=self.table_name,
            vec_data=(embedding if not self.normalize else _normalize(embedding)),
            vec_column_name=self.vector_field,
            distance_func=self._parse_metric_type_str_to_dist_func(),
            topk=k,
            output_column_names=[self.text_field, self.metadata_field],
            where_clause=([text(filter)] if filter is not None else None),
        )
        return [
            Chunk(
                content=r[0],
                metadata=json.loads(r[1]),
            )
            for r in res.fetchall()
        ]

    def _similarity_search_with_score_by_vector(
        self,
        embedding: List[float],
        k: int = 10,
        param: Optional[dict] = None,
        filters: Optional[MetadataFilters] = None,
        **kwargs: Any,
    ) -> List[Tuple[Chunk, str, float]]:
        if filters is not None:
            filter = self._convert_metadata_filters(filters)
        else:
            filter = None

        search_param = (
            param if param is not None else DEFAULT_OCEANBASE_HNSW_SEARCH_PARAM
        )
        ef_search = search_param.get(
            "efSearch", DEFAULT_OCEANBASE_HNSW_SEARCH_PARAM["efSearch"]
        )
        if ef_search != self.hnsw_ef_search:
            self.vector_store_client.set_ob_hnsw_ef_search(ef_search)
            self.hnsw_ef_search = ef_search

        res = self.vector_store_client.ann_search(
            table_name=self.table_name,
            vec_data=(embedding if not self.normalize else _normalize(embedding)),
            vec_column_name=self.vector_field,
            distance_func=self._parse_metric_type_str_to_dist_func(),
            with_dist=True,
            topk=k,
            output_column_names=[
                self.text_field,
                self.metadata_field,
                self.primary_field,
            ],
            where_clause=([text(filter)] if filter is not None else None),
            **kwargs,
        )
        return [
            (
                Chunk(content=r[0], metadata=json.loads(r[1])),
                r[2],
                r[3],
            )
            for r in res.fetchall()
        ]

    def delete_vector_name(self, vector_name: str):
        """Delete vector name."""
        self.vector_store_client.drop_table_if_exist(table_name=self.table_name)

    def delete_by_ids(self, ids: str):
        """Delete vector by ids."""
        split_ids = ids.split(",")
        self.vector_store_client.delete(table_name=self.table_name, ids=split_ids)

    def _enhance_filter_key(self, filter_key: str) -> str:
        return f"{self.metadata_field}->'$.{filter_key}'"

    def _convert_metadata_filters(self, metafilters: MetadataFilters) -> str:
        filters = []
        for filter in metafilters.filters:
            filter_value = _parse_filter_value(filter.value)

            if filter.operator == FilterOperator.EQ:
                filters.append(f"{self._enhance_filter_key(filter.key)}={filter_value}")
            elif filter.operator == FilterOperator.GT:
                filters.append(f"{self._enhance_filter_key(filter.key)}>{filter_value}")
            elif filter.operator == FilterOperator.LT:
                filters.append(f"{self._enhance_filter_key(filter.key)}<{filter_value}")
            elif filter.operator == FilterOperator.NE:
                filters.append(
                    f"{self._enhance_filter_key(filter.key)}!={filter_value}"
                )
            elif filter.operator == FilterOperator.GTE:
                filters.append(
                    f"{self._enhance_filter_key(filter.key)}>={filter_value}"
                )
            elif filter.operator == FilterOperator.LTE:
                filters.append(
                    f"{self._enhance_filter_key(filter.key)}<={filter_value}"
                )
            elif filter.operator == FilterOperator.IN:
                filters.append(
                    f"{self._enhance_filter_key(filter.key)} in {filter_value}"
                )
            elif filter.operator == FilterOperator.NIN:
                filters.append(
                    f"{self._enhance_filter_key(filter.key)} not in {filter_value}"
                )
            else:
                raise ValueError(
                    f"Operator {filter.operator} ('{filter.operator.value}') "
                    f"is not supported by OceanBase."
                )
        return f" {metafilters.condition.value} ".join(filters)



"""
Run unit test with command: pytest dbgpt/datasource/rdbms/tests/test_conn_sqlite.py
"""

import os
import tempfile

import pytest

from dbgpt_ext.datasource.rdbms.conn_sqlite import SQLiteConnector


@pytest.fixture
def db():
    temp_db_file = tempfile.NamedTemporaryFile(delete=False)
    temp_db_file.close()
    conn = SQLiteConnector.from_file_path(temp_db_file.name)
    yield conn
    try:
        # TODO: Failed on windows
        os.unlink(temp_db_file.name)
    except Exception as e:
        print(f"An error occurred: {e}")


def test_get_table_names(db):
    assert list(db.get_table_names()) == []


def test_get_table_info(db):
    assert db.get_table_info() == ""


def test_get_table_info_with_table(db):
    db.run("CREATE TABLE test (id INTEGER);")
    print(db._sync_tables_from_db())
    table_info = db.get_table_info()
    assert "CREATE TABLE test" in table_info


def test_run_sql(db):
    result = db.run("CREATE TABLE test(id INTEGER);")
    assert result[0] == ("id", "INTEGER", 0, None, 0)


def test_run_no_throw(db):
    assert db.run_no_throw("this is a error sql") == []


def test_get_indexes(db):
    db.run("CREATE TABLE test (name TEXT);")
    db.run("CREATE INDEX idx_name ON test(name);")
    indexes = db.get_indexes("test")
    assert indexes == [{"name": "idx_name", "column_names": ["name"]}]


def test_get_indexes_empty(db):
    db.run("CREATE TABLE test (id INTEGER PRIMARY KEY);")
    assert db.get_indexes("test") == []


def test_get_show_create_table(db):
    db.run("CREATE TABLE test (id INTEGER PRIMARY KEY);")
    assert (
        db.get_show_create_table("test") == "CREATE TABLE test (id INTEGER PRIMARY KEY)"
    )


def test_get_fields(db):
    db.run("CREATE TABLE test (id INTEGER PRIMARY KEY);")
    assert db.get_fields("test") == [("id", "INTEGER", 0, None, 1)]


def test_get_charset(db):
    assert db.get_charset() == "UTF-8"


def test_get_collation(db):
    assert db.get_collation() == "UTF-8"


def test_table_simple_info(db):
    db.run("CREATE TABLE test (id INTEGER PRIMARY KEY);")
    assert db.table_simple_info() == ["test(id);"]


def test_get_table_info_no_throw(db):
    db.run("CREATE TABLE test (id INTEGER PRIMARY KEY);")
    assert db.get_table_info_no_throw("xxxx_table").startswith("Error:")


def test_query_ex(db):
    db.run("CREATE TABLE test (id INTEGER PRIMARY KEY);")
    db.run("insert into test(id) values (1)")
    db.run("insert into test(id) values (2)")
    field_names, result = db.query_ex("select * from test")
    assert field_names == ["id"]
    assert result == [(1,), (2,)]

    field_names, result = db.query_ex("select * from test", fetch="one")
    assert field_names == ["id"]
    assert result == [1]


def test_convert_sql_write_to_select(db):
    # TODO
    pass


def test_get_grants(db):
    assert db.get_grants() == []


def test_get_users(db):
    assert db.get_users() == []


def test_get_table_comments(db):
    assert db.get_table_comments() == []
    db.run("CREATE TABLE test (id INTEGER PRIMARY KEY);")
    assert db.get_table_comments() == [
        ("test", "CREATE TABLE test (id INTEGER PRIMARY KEY)")
    ]


def test_get_database_names(db):
    db.get_database_names() == []


def test_db_dir_exist_dir():
    with tempfile.TemporaryDirectory() as temp_dir:
        new_dir = os.path.join(temp_dir, "new_dir")
        file_path = os.path.join(new_dir, "sqlite.db")
        db = SQLiteConnector.from_file_path(file_path)
        assert os.path.exists(new_dir) is True
        assert list(db.get_table_names()) == []
    with tempfile.TemporaryDirectory() as existing_dir:
        file_path = os.path.join(existing_dir, "sqlite.db")
        db = SQLiteConnector.from_file_path(file_path)
        assert os.path.exists(existing_dir) is True
        assert list(db.get_table_names()) == []



"""Neo4j Community Store Adapter."""

import json
import logging
from typing import Any, AsyncGenerator, Dict, Iterator, List, Optional, Tuple, Union

from dbgpt.storage.graph_store.graph import (
    Direction,
    Edge,
    Graph,
    GraphElemType,
    MemoryGraph,
    Vertex,
)
from dbgpt.storage.knowledge_graph.base import ParagraphChunk
from dbgpt_ext.storage.graph_store.neo4j_store import Neo4jStore
from dbgpt_ext.storage.knowledge_graph.community.base import (
    Community,
    GraphStoreAdapter,
)

logger = logging.getLogger(__name__)


class Neo4jStoreAdapter(GraphStoreAdapter):
    """Neo4j Community Store Adapter."""

    MAX_QUERY_LIMIT = 1000
    MAX_HIERARCHY_LEVEL = 3

    def __init__(self, graph_store: Neo4jStore):
        """Initialize Neo4j Community Store Adapter."""
        super().__init__(graph_store)

        # Create indexes for better performance
        self.create_graph(self.graph_store.get_config().name)

    async def discover_communities(self, **kwargs) -> List[str]:
        """Run community discovery.

        Note: Neo4j Graph Data Science library would be optimal for this.
        For now, we return communities based on existing _community_id values.
        """
        query = f"""
        MATCH (n:{self.get_vertex_type()})
        WHERE n._community_id IS NOT NULL
        RETURN DISTINCT n._community_id as community_id
        """
        result = self.graph_store.conn.run(query)
        community_ids = [
            record["community_id"] for record in result if record.get("community_id")
        ]
        logger.info(f"Discovered {len(community_ids)} communities.")
        return community_ids

    async def get_community(self, community_id: str) -> Community:
        """Get community."""
        query = (
            f"MATCH (n:{self.get_vertex_type()}) "
            f"WHERE n._community_id = $community_id "
            f"RETURN n"
        )
        edge_query = (
            f"MATCH (n:{self.get_vertex_type()})-"
            f"[r:{self.get_edge_type()}]-"
            f"(m:{self.get_vertex_type()}) "
            f"WHERE n._community_id = $community_id "
            f"RETURN n, r, m"
        )

        all_vertex_graph = self.query(query, community_id=community_id)
        all_edge_graph = self.query(edge_query, community_id=community_id)
        all_graph = MemoryGraph()
        for vertex in all_vertex_graph.vertices():
            all_graph.upsert_vertex(vertex)
        for edge in all_edge_graph.edges():
            all_graph.append_edge(edge)

        return Community(id=community_id, data=all_graph)

    @property
    def graph_store(self) -> Neo4jStore:
        """Get the graph store."""
        return self._graph_store  # type: ignore[return-value]

    def get_graph_config(self):
        """Get the graph store config."""
        return self.graph_store.get_config()

    def get_vertex_type(self) -> str:
        """Get the vertex type."""
        return GraphElemType.ENTITY.value

    def get_edge_type(self) -> str:
        """Get the edge type."""
        return GraphElemType.RELATION.value

    def get_triplets(self, subj: str) -> List[Tuple[str, str]]:
        """Get triplets."""
        triplet_query = (
            f"MATCH (n1:{GraphElemType.ENTITY.value})-[r]->(n2:"
            f"{GraphElemType.ENTITY.value}) "
            f"WHERE n1.id = $subj "
            f"RETURN type(r) as rel, n2.id as obj"
        )
        data = self.graph_store.conn.run(triplet_query, subj=subj)
        return [(record["rel"], record["obj"]) for record in data]

    def get_document_vertex(self, doc_name: str) -> Vertex:
        """Get the document vertex in the graph."""
        query = f"""MATCH (n:{GraphElemType.DOCUMENT.value}) 
        WHERE n.name = $doc_name RETURN n LIMIT 1"""
        graph = self.query(query, doc_name=doc_name)
        if graph.vertex_count > 0:
            return list(graph.vertices())[0]
        return None

    async def get_schema(self, refresh: bool = False) -> str:
        """Get the schema of the graph store."""
        try:
            # Get node labels
            node_query = "CALL db.labels()"
            labels = [
                record["label"] for record in self.graph_store.conn.run(node_query)
            ]

            # Get relationship types
            rel_query = "CALL db.relationshipTypes()"
            rel_types = [
                record["relationshipType"]
                for record in self.graph_store.conn.run(rel_query)
            ]

            schema = {
                "node_labels": labels,
                "relationship_types": rel_types,
            }

            return json.dumps(schema, indent=2)
        except Exception as e:
            logger.error(f"Failed to get schema: {e}")
            return "{}"

    def get_full_graph(self, limit: Optional[int] = None) -> Graph:
        """Get full graph."""
        if not limit:
            limit = self.MAX_QUERY_LIMIT
        if limit <= 0:
            raise ValueError("Limit must be greater than 0.")

        graph_result = self.query(f"MATCH (n)-[r]-(m) RETURN n, r, m LIMIT {limit}")
        full_graph = MemoryGraph()
        for vertex in graph_result.vertices():
            full_graph.upsert_vertex(vertex)
        for edge in graph_result.edges():
            full_graph.append_edge(edge)
        return full_graph

    def _escape_quotes(self, value: Any) -> str:
        """Escape quotes in string values."""
        if value is None:
            return ""
        return str(value).replace("'", "\\'").replace('"', '\\"')

    def upsert_entities(self, entities: Iterator[Vertex]) -> None:
        """Upsert entities."""
        for entity in entities:
            props = {
                "id": entity.vid,
                "name": entity.name or entity.vid,
                "description": entity.get_prop("description") or "",
                "_document_id": entity.get_prop("_document_id") or "0",
                "_chunk_id": entity.get_prop("_chunk_id") or "0",
                "_community_id": entity.get_prop("_community_id") or "0",
            }

            query = f"""
            MERGE (n:{GraphElemType.ENTITY.value} {{id: $id}})
            SET n += $props
            """
            self.graph_store.conn.run(query, id=entity.vid, props=props)

    def upsert_edge(
        self, edges: Iterator[Edge], edge_type: str, src_type: str, dst_type: str
    ) -> None:
        """Upsert edges."""
        for edge in edges:
            props = {
                "id": edge.name or f"{edge.sid}_{edge.tid}",
                "name": edge.name or edge_type,
                "description": edge.get_prop("description") or "",
                "_chunk_id": edge.get_prop("_chunk_id") or "",
            }

            query = f"""
            MATCH (src:{src_type} {{id: $sid}})
            MATCH (dst:{dst_type} {{id: $tid}})
            MERGE (src)-[r:{edge_type}]->(dst)
            SET r += $props
            """
            self.graph_store.conn.run(query, sid=edge.sid, tid=edge.tid, props=props)

    def upsert_chunks(self, chunks: Iterator[Union[Vertex, ParagraphChunk]]) -> None:
        """Upsert chunks."""
        for chunk in chunks:
            if isinstance(chunk, ParagraphChunk):
                props = {
                    "id": chunk.chunk_id,
                    "name": chunk.chunk_name or chunk.chunk_id,
                    "content": chunk.content or "",
                }
            else:
                props = {
                    "id": chunk.vid,
                    "name": chunk.name or chunk.vid,
                    "content": chunk.get_prop("content") or "",
                }

            query = f"""
            MERGE (n:{GraphElemType.CHUNK.value} {{id: $id}})
            SET n += $props
            """
            self.graph_store.conn.run(query, id=props["id"], props=props)

    def upsert_documents(
        self, documents: Iterator[Union[Vertex, ParagraphChunk]]
    ) -> None:
        """Upsert documents."""
        for document in documents:
            if isinstance(document, ParagraphChunk):
                props = {
                    "id": document.chunk_id,
                    "name": document.chunk_name or document.chunk_id,
                    "content": "",
                }
            else:
                props = {
                    "id": document.vid,
                    "name": document.name or document.vid,
                    "content": "",
                }

            query = f"""
            MERGE (n:{GraphElemType.DOCUMENT.value} {{id: $id}})
            SET n += $props
            """
            self.graph_store.conn.run(query, id=props["id"], props=props)

    def insert_triplet(self, subj: str, rel: str, obj: str) -> None:
        """Add triplet."""
        # Create subject and object entities
        query = f"""
        MERGE (s:{GraphElemType.ENTITY.value} {{id: $subj}})
        SET s.name = $subj
        MERGE (o:{GraphElemType.ENTITY.value} {{id: $obj}})
        SET o.name = $obj
        """
        self.graph_store.conn.run(query, subj=subj, obj=obj)

        # Create relationship
        edge_query = f"""
        MATCH (s:{GraphElemType.ENTITY.value} {{id: $subj}})
        MATCH (o:{GraphElemType.ENTITY.value} {{id: $obj}})
        MERGE (s)-[r:{GraphElemType.RELATION.value}]->(o)
        SET r.id = $rel, r.name = $rel
        """
        self.graph_store.conn.run(edge_query, subj=subj, obj=obj, rel=rel)

    def upsert_graph(self, graph: Graph) -> None:
        """Add graph to the graph store."""
        # Get the iterators of all the vertices and the edges from the graph
        documents: Iterator[Vertex] = graph.vertices(
            filter_fn=lambda x: (
                x.get_prop("vertex_type") == GraphElemType.DOCUMENT.value
            )
        )
        chunks: Iterator[Vertex] = graph.vertices(
            filter_fn=lambda x: x.get_prop("vertex_type") == GraphElemType.CHUNK.value
        )
        entities: Iterator[Vertex] = graph.vertices(
            filter_fn=lambda x: x.get_prop("vertex_type") == GraphElemType.ENTITY.value
        )
        doc_include_chunk: Iterator[Edge] = graph.edges(
            filter_fn=lambda x: (
                x.get_prop("edge_type") == GraphElemType.DOCUMENT_INCLUDE_CHUNK.value
            )
        )
        chunk_include_chunk: Iterator[Edge] = graph.edges(
            filter_fn=lambda x: (
                x.get_prop("edge_type") == GraphElemType.CHUNK_INCLUDE_CHUNK.value
            )
        )
        chunk_include_entity: Iterator[Edge] = graph.edges(
            filter_fn=lambda x: (
                x.get_prop("edge_type") == GraphElemType.CHUNK_INCLUDE_ENTITY.value
            )
        )
        chunk_next_chunk: Iterator[Edge] = graph.edges(
            filter_fn=lambda x: (
                x.get_prop("edge_type") == GraphElemType.CHUNK_NEXT_CHUNK.value
            )
        )
        relation: Iterator[Edge] = graph.edges(
            filter_fn=lambda x: x.get_prop("edge_type") == GraphElemType.RELATION.value
        )

        # Upsert the vertices and the edges to the graph store
        self.upsert_entities(entities)
        self.upsert_chunks(chunks)
        self.upsert_documents(documents)
        self.upsert_edge(
            doc_include_chunk,
            GraphElemType.INCLUDE.value,
            GraphElemType.DOCUMENT.value,
            GraphElemType.CHUNK.value,
        )
        self.upsert_edge(
            chunk_include_chunk,
            GraphElemType.INCLUDE.value,
            GraphElemType.CHUNK.value,
            GraphElemType.CHUNK.value,
        )
        self.upsert_edge(
            chunk_include_entity,
            GraphElemType.INCLUDE.value,
            GraphElemType.CHUNK.value,
            GraphElemType.ENTITY.value,
        )
        self.upsert_edge(
            chunk_next_chunk,
            GraphElemType.NEXT.value,
            GraphElemType.CHUNK.value,
            GraphElemType.CHUNK.value,
        )
        self.upsert_edge(
            relation,
            GraphElemType.RELATION.value,
            GraphElemType.ENTITY.value,
            GraphElemType.ENTITY.value,
        )

    def upsert_doc_include_chunk(self, chunk: ParagraphChunk) -> None:
        """Convert chunk to document include chunk."""
        # Create document vertex
        doc_id = (
            f"doc_{chunk.chunk_name}" if chunk.chunk_name else f"doc_{chunk.chunk_id}"
        )
        doc_query = f"""
        MERGE (d:{GraphElemType.DOCUMENT.value} {{id: $doc_id}})
        SET d.name = $doc_name
        """
        self.graph_store.conn.run(
            query=doc_query, doc_id=doc_id, doc_name=chunk.chunk_name or doc_id
        )

        # Create chunk vertex
        chunk_query = f"""
        MERGE (c:{GraphElemType.CHUNK.value} {{id: $chunk_id}})
        SET c.name = $chunk_name, c.content = $content
        """
        self.graph_store.conn.run(
            query=chunk_query,
            chunk_id=chunk.chunk_id,
            chunk_name=chunk.chunk_name or chunk.chunk_id,
            content=chunk.content or "",
        )

        # Create relationship
        rel_query = f"""
        MATCH (d:{GraphElemType.DOCUMENT.value} {{id: $doc_id}})
        MATCH (c:{GraphElemType.CHUNK.value} {{id: $chunk_id}})
        MERGE (d)-[r:{GraphElemType.INCLUDE.value}]->(c)
        SET r.id = $rel_id, r.name = 'include'
        """
        self.graph_store.conn.run(
            query=rel_query,
            doc_id=doc_id,
            chunk_id=chunk.chunk_id,
            rel_id=f"{doc_id}_{chunk.chunk_id}",
        )

    def upsert_chunk_include_chunk(self, chunk: ParagraphChunk) -> None:
        """Convert chunk to chunk include chunk."""
        # This is typically used for hierarchical chunks
        # For now, we just ensure the chunk exists
        chunk_query = f"""
        MERGE (c:{GraphElemType.CHUNK.value} {{id: $chunk_id}})
        SET c.name = $chunk_name, c.content = $content
        """
        self.graph_store.conn.run(
            query=chunk_query,
            chunk_id=chunk.chunk_id,
            chunk_name=chunk.chunk_name or chunk.chunk_id,
            content=chunk.content or "",
        )

    def upsert_chunk_next_chunk(
        self, chunk: ParagraphChunk, next_chunk: ParagraphChunk
    ):
        """Upsert the vertices and the edge in chunk_next_chunk."""
        # Create both chunks
        for c in [chunk, next_chunk]:
            chunk_query = f"""
            MERGE (c:{GraphElemType.CHUNK.value} {{id: $chunk_id}})
            SET c.name = $chunk_name, c.content = $content
            """
            self.graph_store.conn.run(
                query=chunk_query,
                chunk_id=c.chunk_id,
                chunk_name=c.chunk_name or c.chunk_id,
                content=c.content or "",
            )

        # Create NEXT relationship
        rel_query = f"""
        MATCH (c1:{GraphElemType.CHUNK.value} {{id: $chunk_id1}})
        MATCH (c2:{GraphElemType.CHUNK.value} {{id: $chunk_id2}})
        MERGE (c1)-[r:{GraphElemType.NEXT.value}]->(c2)
        SET r.id = $rel_id, r.name = 'next'
        """
        self.graph_store.conn.run(
            query=rel_query,
            chunk_id1=chunk.chunk_id,
            chunk_id2=next_chunk.chunk_id,
            rel_id=f"{chunk.chunk_id}_{next_chunk.chunk_id}",
        )

    def upsert_chunk_include_entity(
        self, chunk: ParagraphChunk, entity: Vertex
    ) -> None:
        """Convert chunk to chunk include entity."""
        # Create chunk
        chunk_query = f"""
        MERGE (c:{GraphElemType.CHUNK.value} {{id: $chunk_id}})
        SET c.name = $chunk_name, c.content = $content
        """
        self.graph_store.conn.run(
            query=chunk_query,
            chunk_id=chunk.chunk_id,
            chunk_name=chunk.chunk_name or chunk.chunk_id,
            content=chunk.content or "",
        )

        # Create entity
        entity_query = f"""
        MERGE (e:{GraphElemType.ENTITY.value} {{id: $entity_id}})
        SET e.name = $entity_name
        """
        self.graph_store.conn.run(
            query=entity_query,
            entity_id=entity.vid,
            entity_name=entity.name or entity.vid,
        )

        # Create relationship
        rel_query = f"""
        MATCH (c:{GraphElemType.CHUNK.value} {{id: $chunk_id}})
        MATCH (e:{GraphElemType.ENTITY.value} {{id: $entity_id}})
        MERGE (c)-[r:{GraphElemType.INCLUDE.value}]->(e)
        SET r.id = $rel_id, r.name = 'include'
        """
        self.graph_store.conn.run(
            query=rel_query,
            chunk_id=chunk.chunk_id,
            entity_id=entity.vid,
            rel_id=f"{chunk.chunk_id}_{entity.vid}",
        )

    def delete_document(self, chunk_id: str) -> None:
        """Delete document in the graph."""
        chunkids_list = [c.strip() for c in chunk_id.split(",")]

        # Delete chunks
        del_chunk_query = f"""
        MATCH (n:{GraphElemType.CHUNK.value})
        WHERE n.id IN $chunk_ids
        DETACH DELETE n
        """
        self.graph_store.conn.run(del_chunk_query, chunk_ids=chunkids_list)

        # Delete relations related to these chunks
        del_relation_query = f"""
        MATCH (m:{GraphElemType.ENTITY.value})-[r:{GraphElemType.RELATION.value}]
        -(n:{GraphElemType.ENTITY.value})
        WHERE r._chunk_id IN $chunk_ids
        DELETE r
        """
        self.graph_store.conn.run(del_relation_query, chunk_ids=chunkids_list)

        # Delete orphan nodes
        delete_orphan_query = """
        MATCH (n)
        WHERE NOT EXISTS((n)-[]-())
        DELETE n
        """
        self.graph_store.conn.run(delete_orphan_query)

    def delete_triplet(self, sub: str, rel: str, obj: str) -> None:
        """Delete triplet."""
        del_query = f"""
        MATCH (n1:{GraphElemType.ENTITY.value} {{id: $sub}})
        -[r:{GraphElemType.RELATION.value} {{id: $rel}}]->
        (n2:{GraphElemType.ENTITY.value} {{id: $obj}})
        DELETE r
        """
        self.graph_store.conn.run(query=del_query, sub=sub, rel=rel, obj=obj)

    def drop(self):
        """Delete Graph."""
        self.truncate()

    def create_graph(self, graph_name: str):
        """Create indexes for the graph.

        Note: In Neo4j, we don't create separate graphs like in TuGraph.
        All data is stored in the configured database, and we use labels
        to organize different knowledge spaces.
        """
        logger.info(f"Setting up Neo4j for knowledge space: {graph_name}")

        # Create indexes for common properties
        # Note: These run on the database specified in the connection config
        indexes = [
            f"CREATE INDEX IF NOT EXISTS FOR (n:{GraphElemType.ENTITY.value}) ON"
            f" (n.id)",
            f"CREATE INDEX IF NOT EXISTS FOR (n:{GraphElemType.ENTITY.value}) ON"
            f" (n._community_id)",
            f"CREATE INDEX IF NOT EXISTS FOR (n:{GraphElemType.CHUNK.value}) ON (n.id)",
            f"CREATE INDEX IF NOT EXISTS FOR (n:{GraphElemType.CHUNK.value}) ON"
            f" (n.content)",
            f"CREATE INDEX IF NOT EXISTS FOR (n:{GraphElemType.DOCUMENT.value}) ON"
            f" (n.id)",
            f"CREATE INDEX IF NOT EXISTS FOR (n:{GraphElemType.DOCUMENT.value}) ON"
            f" (n.name)",
        ]

        for index_query in indexes:
            try:
                self.graph_store.conn.run(index_query)
                logger.info("Successfully created/verified index")
            except Exception as e:
                # Log as info instead of warning - indexes might already exist
                logger.info(f"Index creation note: {str(e)[:200]}")
                # Continue even if index creation fails - it might already exist

    def create_graph_label(
        self,
        graph_elem_type: GraphElemType,
        graph_properties: List[Dict[str, Union[str, bool]]],
    ) -> None:
        """Create a graph label.

        Neo4j creates labels dynamically, so we just log this for compatibility.
        """
        logger.info(f"Neo4j: Label {graph_elem_type.value} will be created dynamically")

    def truncate(self):
        """Truncate Graph."""
        logger.warning("Truncating all data from Neo4j database")

        # Delete all relationships first
        self.graph_store.conn.run("MATCH ()-[r]->() DELETE r")

        # Then delete all nodes
        self.graph_store.conn.run("MATCH (n) DELETE n")

    def check_label(self, graph_elem_type: GraphElemType) -> bool:
        """Check if the label exists in the graph."""
        try:
            query = "CALL db.labels()"
            labels = [record[0] for record in self.graph_store.conn.run(query)]
            return graph_elem_type.value in labels
        except Exception as e:
            logger.warning(f"Failed to check label: {e}")
            return False

    def explore_trigraph(
        self,
        subs: Union[List[str], List[List[float]]],
        topk: Optional[int] = None,
        score_threshold: Optional[float] = None,
        direct: Direction = Direction.BOTH,
        depth: int = 3,
        fan: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> MemoryGraph:
        """Explore the graph from given subjects up to a depth."""
        if not subs:
            return MemoryGraph()

        if depth <= 0:
            depth = 3

        if limit is None:
            limit = self.MAX_QUERY_LIMIT

        # Convert direction
        if direct == Direction.OUT:
            direction_str = ">"
        elif direct == Direction.IN:
            direction_str = "<"
        else:
            direction_str = ""

        # Build query for each subject
        result_graph = MemoryGraph()

        for sub in subs:
            if isinstance(sub, list):
                # Embedding search not yet implemented for Neo4j
                continue

            query = f"""
            MATCH path = (start:{GraphElemType.ENTITY.value} {{id: $sub}})
            -[*1..{depth}]-{direction_str}(end)
            RETURN nodes(path) as nodes, relationships(path) as rels
            LIMIT {limit}
            """

            try:
                results = self.graph_store.conn.run(query, sub=sub)
                for record in results:
                    # Process nodes
                    for node in record.get("nodes", []):
                        vertex = self._neo4j_node_to_vertex(node)
                        result_graph.upsert_vertex(vertex)

                    # Process relationships
                    for rel in record.get("rels", []):
                        edge = self._neo4j_relationship_to_edge(rel)
                        result_graph.append_edge(edge)
            except Exception as e:
                logger.error(f"Failed to explore trigraph: {e}")

        return result_graph

    def explore_docgraph_with_entities(
        self,
        subs: List[str],
        topk: Optional[int] = None,
        score_threshold: Optional[float] = None,
        direct: Direction = Direction.BOTH,
        depth: int = 3,
        fan: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> MemoryGraph:
        """Explore the graph from given subjects up to a depth."""
        if not subs:
            return MemoryGraph()

        if limit is None:
            limit = self.MAX_QUERY_LIMIT

        result_graph = MemoryGraph()

        for sub in subs:
            # Find chunks connected to this entity
            query = f"""
            MATCH (e:{GraphElemType.ENTITY.value} {{id: $sub}})
            <-[:{GraphElemType.INCLUDE.value}]-
            (c:{GraphElemType.CHUNK.value})
            <-[:{GraphElemType.INCLUDE.value}]-
            (d:{GraphElemType.DOCUMENT.value})
            RETURN e, c, d
            LIMIT {limit}
            """

            try:
                results = self.graph_store.conn.run(query, sub=sub)
                for record in results:
                    for key in ["e", "c", "d"]:
                        if key in record and record[key]:
                            vertex = self._neo4j_node_to_vertex(record[key])
                            result_graph.upsert_vertex(vertex)
            except Exception as e:
                logger.error(f"Failed to explore docgraph: {e}")

        return result_graph

    def explore_docgraph_without_entities(
        self,
        subs: Union[List[str], List[List[float]]],
        topk: Optional[int] = None,
        score_threshold: Optional[float] = None,
        direct: Direction = Direction.BOTH,
        depth: int = 3,
        fan: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> MemoryGraph:
        """Explore the graph from given subjects up to a depth."""
        if not subs:
            return MemoryGraph()

        if limit is None:
            limit = self.MAX_QUERY_LIMIT

        result_graph = MemoryGraph()

        for sub in subs:
            if isinstance(sub, list):
                # Embedding search not yet implemented
                continue

            # Find chunks containing the keyword
            query = f"""
            MATCH (c:{GraphElemType.CHUNK.value})
            WHERE c.content CONTAINS $keyword
            MATCH (d:{GraphElemType.DOCUMENT.value})
            -[:{GraphElemType.INCLUDE.value}]->(c)
            RETURN c, d
            LIMIT {limit}
            """

            try:
                results = self.graph_store.conn.run(query, keyword=sub)
                for record in results:
                    for key in ["c", "d"]:
                        if key in record and record[key]:
                            vertex = self._neo4j_node_to_vertex(record[key])
                            result_graph.upsert_vertex(vertex)
            except Exception as e:
                logger.error(f"Failed to explore docgraph: {e}")

        return result_graph

    def query(self, query: str, **kwargs) -> MemoryGraph:
        """Execute a Cypher query and return results as a MemoryGraph."""
        graph = MemoryGraph()

        try:
            results = self.graph_store.conn.run(query, **kwargs)

            for record in results:
                # Process each value in the record
                for key, value in record.items():
                    if value is None:
                        continue
                    if hasattr(value, "labels"):  # It's a node
                        vertex = self._neo4j_node_to_vertex(value)
                        graph.upsert_vertex(vertex)
                    elif hasattr(value, "type"):  # It's a relationship
                        edge = self._neo4j_relationship_to_edge(value)
                        graph.append_edge(edge)

        except Exception as e:
            logger.error(f"Query execution failed: {e}\nQuery: {query}")

        return graph

    async def stream_query(self, query: str, **kwargs) -> AsyncGenerator[Graph, None]:
        """Execute a stream query."""
        # For now, return the full result as a single graph
        result = self.query(query, **kwargs)
        yield result

    def _neo4j_node_to_vertex(self, node) -> Vertex:
        """Convert a Neo4j node to a Vertex."""
        props = dict(node.items())
        vertex_id = props.get("id", str(node.id))
        labels = list(node.labels)

        if labels:
            props["vertex_type"] = labels[0]

        return Vertex(vid=vertex_id, props=props)

    def _neo4j_relationship_to_edge(self, rel) -> Edge:
        """Convert a Neo4j relationship to an Edge."""
        props = dict(rel.items())
        props["edge_type"] = rel.type

        # Get source and target node custom IDs (not internal Neo4j IDs)
        # Nodes are stored with custom 'id' property, must use that for consistency
        start_node = rel.start_node if hasattr(rel, "start_node") else rel.nodes[0]
        end_node = rel.end_node if hasattr(rel, "end_node") else rel.nodes[1]

        # Extract custom 'id' property from nodes, fallback to internal id if missing
        sid = str(dict(start_node.items()).get("id", start_node.id))
        tid = str(dict(end_node.items()).get("id", end_node.id))

        # Edge requires name as a positional parameter
        # Remove 'name' from props to avoid passing it twice
        edge_name = props.pop("name", rel.type)
        return Edge(sid, tid, edge_name, **props)



"""Graph Retriever."""

import logging
import os
from typing import List, Optional, Tuple, Union

from dbgpt.core import Embeddings, LLMClient
from dbgpt.rag.transformer.keyword_extractor import KeywordExtractor
from dbgpt.rag.transformer.simple_intent_translator import SimpleIntentTranslator
from dbgpt.storage.graph_store.graph import Graph, MemoryGraph
from dbgpt_ext.rag.transformer.local_text2gql import LocalText2GQL
from dbgpt_ext.rag.transformer.text2gql import Text2GQL

from ...transformer.text_embedder import TextEmbedder
from .base import GraphRetrieverBase
from .document_graph_retriever import (
    DocumentGraphRetriever,
)
from .keyword_based_graph_retriever import (  # noqa: E501
    KeywordBasedGraphRetriever,
)
from .text_based_graph_retriever import (
    TextBasedGraphRetriever,
)
from .vector_based_graph_retriever import (
    VectorBasedGraphRetriever,
)

logger = logging.getLogger(__name__)


class GraphRetriever(GraphRetrieverBase):
    """Graph Retriever class."""

    def __init__(
        self,
        graph_store_adapter,
        llm_client: Optional[LLMClient] = None,
        llm_model: Optional[str] = None,
        triplet_graph_enabled: Optional[bool] = True,
        document_graph_enabled: Optional[bool] = True,
        extract_top_k: Optional[int] = 5,
        kg_chunk_search_top_k: Optional[int] = 5,
        similarity_top_k: Optional[int] = 5,
        similarity_score_threshold: Optional[float] = 0.7,
        embedding_fn: Optional[Embeddings] = None,
        embedding_batch_size: Optional[int] = 20,
        enable_text_search: Optional[bool] = False,
        text2gql_model_enabled: Optional[bool] = False,
        text2gql_model_name: Optional[str] = None,
    ):
        """Initialize Graph Retriever."""
        self._triplet_graph_enabled = triplet_graph_enabled or (
            os.getenv("TRIPLET_GRAPH_ENABLED", "").lower() == "true"
        )
        self._document_graph_enabled = document_graph_enabled or (
            os.getenv("DOCUMENT_GRAPH_ENABLED", "").lower() == "true"
        )
        triplet_topk = int(
            extract_top_k or os.getenv("KNOWLEDGE_GRAPH_EXTRACT_SEARCH_TOP_SIZE")
        )
        document_topk = int(
            kg_chunk_search_top_k or os.getenv("KNOWLEDGE_GRAPH_CHUNK_SEARCH_TOP_SIZE")
        )
        llm_client = llm_client
        model_name = llm_model
        self._enable_similarity_search = (
            graph_store_adapter.graph_store.enable_similarity_search
        )
        self._embedding_batch_size = int(
            embedding_batch_size or os.getenv("KNOWLEDGE_GRAPH_EMBEDDING_BATCH_SIZE")
        )
        similarity_search_topk = int(
            similarity_top_k or os.getenv("KNOWLEDGE_GRAPH_SIMILARITY_SEARCH_TOP_SIZE")
        )
        similarity_search_score_threshold = float(
            similarity_score_threshold
            or os.getenv("KNOWLEDGE_GRAPH_EXTRACT_SEARCH_RECALL_SCORE")
        )
        self._enable_text_search = enable_text_search or (
            os.getenv("TEXT_SEARCH_ENABLED", "").lower() == "true"
        )
        text2gql_model_enabled = text2gql_model_enabled or (
            os.getenv("TEXT2GQL_MODEL_ENABLED", "").lower() == "true"
        )
        text2gql_model_name = text2gql_model_name or os.getenv("TEXT2GQL_MODEL_NAME")
        text2gql_model_enabled = (
            os.environ["TEXT2GQL_MODEL_ENABLED"].lower() == "true"
            if "TEXT2GQL_MODEL_ENABLED" in os.environ
            else text2gql_model_enabled
        )
        text2gql_model_name = os.getenv(
            "TEXT2GQL_MODEL_NAME",
            text2gql_model_name,
        )
        text2gql_model_enabled = (
            os.environ["TEXT2GQL_MODEL_ENABLED"].lower() == "true"
            if "TEXT2GQL_MODEL_ENABLED" in os.environ
            else text2gql_model_enabled
        )
        text2gql_model_name = os.getenv(
            "TEXT2GQL_MODEL_NAME",
            text2gql_model_name,
        )
        text2gql_model_enabled = (
            os.environ["TEXT2GQL_MODEL_ENABLED"].lower() == "true"
            if "TEXT2GQL_MODEL_ENABLED" in os.environ
            else text2gql_model_enabled
        )
        text2gql_model_name = os.getenv(
            "TEXT2GQL_MODEL_NAME",
            text2gql_model_name,
        )

        self._keyword_extractor = KeywordExtractor(llm_client, model_name)
        self._text_embedder = TextEmbedder(embedding_fn)

        intent_interpreter = SimpleIntentTranslator(llm_client, model_name)
        if text2gql_model_enabled:
            text2gql = LocalText2GQL(text2gql_model_name)
        else:
            text2gql = Text2GQL(llm_client, model_name)

        self._keyword_based_graph_retriever = KeywordBasedGraphRetriever(
            graph_store_adapter, triplet_topk
        )
        self._vector_based_graph_retriever = VectorBasedGraphRetriever(
            graph_store_adapter,
            triplet_topk,
            similarity_search_topk,
            similarity_search_score_threshold,
        )
        self._text_based_graph_retriever = TextBasedGraphRetriever(
            graph_store_adapter,
            triplet_topk,
            intent_interpreter,
            text2gql,
        )
        self._document_graph_retriever = DocumentGraphRetriever(
            graph_store_adapter,
            document_topk,
            similarity_search_topk,
            similarity_search_score_threshold,
        )

    async def retrieve(self, text: str) -> Tuple[Graph, Tuple[Graph, str]]:
        """Retrieve subgraph from triplet graph and document graph."""
        subgraph = MemoryGraph()
        subgraph_for_doc = MemoryGraph()
        text2gql_query = ""

        # Retrieve from triplet graph and document graph
        if self._enable_text_search:
            # Retrieve from knowledge graph with text.
            subgraph, text2gql_query = await self._text_based_graph_retriever.retrieve(
                text
            )

        # Extract keywords from original question
        keywords: List[str] = await self._keyword_extractor.extract(text)

        if subgraph.vertex_count == 0 and subgraph.edge_count == 0:
            # if not enable text search or text search failed to retrieve subgraph

            # Using subs to transfer keywords or embeddings
            subs: Union[List[str], List[List[float]]]

            if self._enable_similarity_search:
                # Embedding the question
                vector = await self._text_embedder.embed(text)
                # Embedding the keywords
                vectors = await self._text_embedder.batch_embed(
                    keywords, batch_size=self._embedding_batch_size
                )
                # Using the embeddings of keywords and question
                vectors.append(vector)
                # Using vectors as subs
                subs = vectors
                logger.info(
                    "Search subgraph with the following keywords and question's "
                    f"embedding vector:\n[KEYWORDS]:{keywords}\n[QUESTION]:{text}"
                )
            else:
                # Using keywords as subs
                subs = keywords
                logger.info(
                    "Search subgraph with the following keywords:\n"
                    f"[KEYWORDS]:{keywords}"
                )

            # If enable triplet graph
            if self._triplet_graph_enabled:
                # Retrieve from triplet graph
                if self._enable_similarity_search:
                    # Retrieve from triplet graph with vectors
                    subgraph = await self._vector_based_graph_retriever.retrieve(subs)
                else:
                    # Retrieve from triplet graph with keywords
                    subgraph = await self._keyword_based_graph_retriever.retrieve(subs)

            # If enable document graph
            if self._document_graph_enabled:
                # Retrieve from document graph
                # If not enable triplet graph or failed to retrieve subgraph
                if subgraph.vertex_count == 0 and subgraph.edge_count == 0:
                    # Using subs to retrieve from document graph
                    subgraph_for_doc = await self._document_graph_retriever.retrieve(
                        subs
                    )
                else:
                    # If retrieve subgraph from triplet graph successfully
                    # Using entities in subgraph to search chunks and doc
                    subgraph_for_doc = await self._document_graph_retriever.retrieve(
                        subgraph
                    )

        return subgraph, (subgraph_for_doc, text2gql_query)



"""GaussDB connector."""

import logging
from dataclasses import dataclass, field
from typing import Any, Iterable, List, Optional, Tuple, Type, cast
from urllib.parse import quote
from urllib.parse import quote_plus as urlquote

from sqlalchemy import text

from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.datasource.rdbms.base import RDBMSConnector, RDBMSDatasourceParameters
from dbgpt.util.i18n_utils import _

logger = logging.getLogger(__name__)


@auto_register_resource(
    label=_("GaussDB datasource"),
    category=ResourceCategory.DATABASE,
    tags={"order": TAGS_ORDER_HIGH},
    description=_(
        "A scalable,enterprise-grade relational database with distributed architecture."
    ),
)
@dataclass
class GaussDBParameters(RDBMSDatasourceParameters):
    """GaussDB connection parameters."""

    __type__ = "gaussdb"
    schema: str = field(
        default="public", metadata={"help": _("Database schema, defaults to 'public'")}
    )
    driver: str = field(
        default="postgresql+psycopg2",
        metadata={
            "help": _("Driver name for postgres, default is postgresql+psycopg2."),
        },
    )

    def create_connector(self) -> "GaussDBConnector":
        """Create GaussDB connector."""
        return GaussDBConnector.from_parameters(self)

    def db_url(self, ssl: bool = False, charset: Optional[str] = None) -> str:
        return f"{self.driver}://{self.user}:{self.password}@{self.host}:{self.port}/{self.database}"


class GaussDBConnector(RDBMSConnector):
    """GaussDBConnector connector."""

    driver = "postgresql+psycopg2"
    db_type = "gaussdb"
    db_dialect = "gaussdb"

    @classmethod
    def param_class(cls) -> Type[GaussDBParameters]:
        """Return the parameter class."""
        return GaussDBParameters

    @classmethod
    def from_uri_db(
        cls,
        host: str,
        port: int,
        user: str,
        pwd: str,
        db_name: str,
        engine_args: Optional[dict] = None,
        **kwargs: Any,
    ) -> "GaussDBConnector":
        """Create a new PostgreSQLConnector from host, port, user, pwd, db_name."""
        db_url: str = (
            f"{cls.driver}://{quote(user)}:{urlquote(pwd)}@{host}:{str(port)}/{db_name}"
        )
        from sqlalchemy.dialects.postgresql.base import PGDialect

        PGDialect._get_server_version_info = lambda *args: (7, 0)
        return cast(GaussDBConnector, cls.from_uri(db_url, engine_args, **kwargs))

    @classmethod
    def from_parameters(cls, parameters: GaussDBParameters) -> "RDBMSConnector":
        """Create a new connector from parameters."""
        return cls.from_uri_db(
            parameters.host,
            parameters.port,
            parameters.user,
            parameters.password,
            parameters.database,
            schema=parameters.schema,
            engine_args=parameters.engine_args(),
        )

    def _sync_tables_from_db(self) -> Iterable[str]:
        """Read table information from database with schema support."""
        schema = self._schema or "public"

        with self.session_scope() as session:
            # Get tables for specific schema
            table_results = session.execute(
                text(
                    """
                    SELECT tablename 
                    FROM pg_catalog.pg_tables 
                    WHERE schemaname = :schema
                    """
                ),
                {"schema": schema},
            )

            # Get views for specific schema
            view_results = session.execute(
                text(
                    """
                    SELECT viewname 
                    FROM pg_catalog.pg_views 
                    WHERE schemaname = :schema
                    """
                ),
                {"schema": schema},
            )

            table_results = set(row[0] for row in table_results)
            view_results = set(row[0] for row in view_results)
            self._all_tables = table_results.union(view_results)

            # Reflect with schema
            self._metadata.reflect(bind=self._engine, schema=schema)
            return self._all_tables

    def get_grants(self):
        """Get grants."""
        with self.session_scope() as session:
            cursor = session.execute(
                text(
                    """
                    SELECT DISTINCT grantee, privilege_type
                    FROM information_schema.role_table_grants
                    WHERE grantee = CURRENT_USER;"""
                )
            )
            grants = cursor.fetchall()
            return grants

    def get_collation(self):
        """Get collation."""
        try:
            with self.session_scope() as session:
                cursor = session.execute(
                    text(
                        "SELECT datcollate AS collation FROM pg_database WHERE "
                        "datname = current_database();"
                    )
                )
                collation = cursor.fetchone()[0]
                return collation
        except Exception as e:
            logger.warning(f"postgresql get collation error: {str(e)}")
            return None

    def get_users(self):
        """Get user info."""
        try:
            with self.session_scope() as session:
                cursor = session.execute(
                    text("SELECT rolname FROM pg_roles WHERE rolname NOT LIKE 'pg_%';")
                )
                users = cursor.fetchall()
                return [user[0] for user in users]
        except Exception as e:
            logger.warning(f"postgresql get users error: {str(e)}")
            return []

    def get_fields(self, table_name: str, db_name: Optional[str] = None) -> List[Tuple]:
        """Get column fields about specified table."""
        schema = self._schema or "public"
        sql = """
            SELECT 
                column_name, 
                data_type, 
                column_default, 
                is_nullable,
                col_description(
                    (quote_ident(:schema) || '.' || quote_ident(:table))::regclass::oid,
                    ordinal_position
                ) as column_comment
            FROM information_schema.columns 
            WHERE table_schema = :schema 
            AND table_name = :table
            ORDER BY ordinal_position
        """
        with self.session_scope() as session:
            cursor = session.execute(
                text(sql),
                {"schema": schema, "table": table_name},
            )
            fields = cursor.fetchall()
            return [
                (field[0], field[1], field[2], field[3], field[4]) for field in fields
            ]

    def get_charset(self):
        """Get character_set."""
        with self.session_scope() as session:
            cursor = session.execute(
                text(
                    "SELECT pg_encoding_to_char(encoding) FROM pg_database WHERE "
                    "datname = current_database();"
                )
            )
            character_set = cursor.fetchone()[0]
            return character_set

    def get_show_create_table(self, table_name: str) -> str:
        """Return show create table with schema support."""
        schema = self._schema or "public"

        with self.session_scope() as session:
            # Get column definitions
            cur = session.execute(
                text(
                    """
                    SELECT 
                        column_name,
                        data_type,
                        column_default,
                        is_nullable,
                        character_maximum_length,
                        numeric_precision,
                        numeric_scale
                    FROM information_schema.columns
                    WHERE table_schema = :schema 
                    AND table_name = :table
                    ORDER BY ordinal_position
                    """
                ),
                {"schema": schema, "table": table_name},
            )

            create_table = f"CREATE TABLE {schema}.{table_name} (\n"
            for row in cur.fetchall():
                col_name = row[0]
                data_type = row[1]
                default = f"DEFAULT {row[2]}" if row[2] else ""
                nullable = "NOT NULL" if row[3] == "NO" else ""

                # Add length/precision/scale if applicable
                if row[4]:  # character_maximum_length
                    data_type = f"{data_type}({row[4]})"
                elif row[5]:  # numeric_precision
                    if row[6]:  # numeric_scale
                        data_type = f"{data_type}({row[5]},{row[6]})"
                    else:
                        data_type = f"{data_type}({row[5]})"

                create_table += f"    {col_name} {data_type} {default} {nullable},\n"

            create_table = create_table.rstrip(",\n") + "\n)"
            return create_table

    def get_table_comments(self, db_name=None):
        """Get table comments."""
        tablses = self.table_simple_info()
        comments = []
        for table in tablses:
            table_name = table[0]
            table_comment = self.get_show_create_table(table_name)
            comments.append((table_name, table_comment))
        return comments

    def get_database_names(self):
        """Get database names."""
        with self.session_scope() as session:
            cursor = session.execute(text("SELECT datname FROM pg_database;"))
            results = cursor.fetchall()
            return [
                d[0]
                for d in results
                if d[0] not in ["template0", "template1", "postgres"]
            ]

    def get_current_db_name(self) -> str:
        """Get current database name."""
        with self.session_scope() as session:
            return session.execute(text("SELECT current_database()")).scalar()

    def table_simple_info(self):
        """Get table simple info."""
        _sql = """
            SELECT table_name, string_agg(column_name, ', ') AS schema_info
            FROM (
                SELECT c.relname AS table_name, a.attname AS column_name
                FROM pg_catalog.pg_class c
                JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
                JOIN pg_catalog.pg_attribute a ON a.attrelid = c.oid
                WHERE c.relkind = 'r'
                AND a.attnum > 0
                AND NOT a.attisdropped
                AND n.nspname NOT LIKE 'pg_%'
                AND n.nspname != 'information_schema'
                ORDER BY c.relname, a.attnum
            ) sub
            GROUP BY table_name;
            """
        with self.session_scope() as session:
            cursor = session.execute(text(_sql))
            results = cursor.fetchall()
            results_str = []
            for result in results:
                results_str.append((str(result[0]), str(result[1])))
            return results_str

    def get_fields_wit_schema(self, table_name, schema_name="public"):
        """Get column fields about specified table."""
        query_sql = f"""
            SELECT c.column_name, c.data_type, c.column_default, c.is_nullable,
             d.description FROM information_schema.columns c
             LEFT JOIN pg_catalog.pg_description d
            ON (c.table_schema || '.' || c.table_name)::regclass::oid = d.objoid
             AND c.ordinal_position = d.objsubid
             WHERE c.table_name='{table_name}' AND c.table_schema='{schema_name}'
        """
        with self.session_scope() as session:
            cursor = session.execute(text(query_sql))
            fields = cursor.fetchall()
            return [
                (field[0], field[1], field[2], field[3], field[4]) for field in fields
            ]

    def get_indexes(self, table_name):
        """Get table indexes about specified table."""
        with self.session_scope() as session:
            cursor = session.execute(
                text(
                    f"SELECT indexname, indexdef FROM pg_indexes WHERE "
                    f"tablename = '{table_name}'"
                )
            )
            indexes = cursor.fetchall()
            return [(index[0], index[1]) for index in indexes]



"""Knowledge Operator."""

from typing import List, Optional

from dbgpt.core import Chunk
from dbgpt.core.awel import MapOperator
from dbgpt.core.awel.flow import (
    IOField,
    OperatorCategory,
    OptionValue,
    Parameter,
    ViewMetadata,
)
from dbgpt.rag.knowledge.base import Knowledge, KnowledgeType
from dbgpt.util.i18n_utils import _
from dbgpt_ext.rag.knowledge.factory import KnowledgeFactory


class KnowledgeOperator(MapOperator[dict, Knowledge]):
    """Knowledge Factory Operator."""

    metadata = ViewMetadata(
        label=_("Knowledge Loader Operator"),
        name="knowledge_operator",
        category=OperatorCategory.RAG,
        description=_(
            _("The knowledge operator, which can create knowledge from datasource.")
        ),
        inputs=[
            IOField.build_from(
                _("knowledge datasource"),
                "knowledge datasource",
                dict,
                _("knowledge datasource, which can be a document, url, or text."),
            )
        ],
        outputs=[
            IOField.build_from(
                _("Knowledge"),
                "Knowledge",
                Knowledge,
                description=_("Knowledge object."),
            )
        ],
        parameters=[
            Parameter.build_from(
                label=_("Default datasource"),
                name="datasource",
                type=str,
                optional=True,
                default=None,
                description=_("Default datasource."),
            ),
            Parameter.build_from(
                label=_("Knowledge type"),
                name="knowledge_type",
                type=str,
                optional=True,
                options=[
                    OptionValue(
                        label="DOCUMENT",
                        name="DOCUMENT",
                        value=KnowledgeType.DOCUMENT.name,
                    ),
                    OptionValue(label="URL", name="URL", value=KnowledgeType.URL.name),
                    OptionValue(
                        label="TEXT", name="TEXT", value=KnowledgeType.TEXT.name
                    ),
                ],
                default=KnowledgeType.DOCUMENT.name,
                description=_("Knowledge type."),
            ),
        ],
        documentation_url="https://github.com/openai/openai-python",
    )

    def __init__(
        self,
        datasource: Optional[str] = None,
        knowledge_type: Optional[str] = KnowledgeType.DOCUMENT.name,
        **kwargs,
    ):
        """Init the query rewrite operator.

        Args:
            knowledge_type: (Optional[KnowledgeType]) The knowledge type.
        """
        super().__init__(**kwargs)
        self._datasource = datasource
        self._knowledge_type = KnowledgeType.get_by_value(knowledge_type)

    async def map(self, datasource: dict) -> Knowledge:
        """Create knowledge from datasource."""
        source = datasource.get("source")
        if self._datasource:
            source = self._datasource
        return await self.blocking_func_to_async(
            KnowledgeFactory.create, source, self._knowledge_type
        )


class ChunksToStringOperator(MapOperator[List[Chunk], str]):
    """The Chunks To String Operator."""

    metadata = ViewMetadata(
        label=_("Chunks To String Operator"),
        name="chunks_to_string_operator",
        description=_("Convert chunks to string."),
        category=OperatorCategory.RAG,
        parameters=[
            Parameter.build_from(
                _("Separator"),
                "separator",
                str,
                description=_("The separator between chunks."),
                optional=True,
                default="\n",
            )
        ],
        inputs=[
            IOField.build_from(
                _("Chunks"),
                "chunks",
                List[Chunk],
                description=_("The input chunks."),
                is_list=True,
            )
        ],
        outputs=[
            IOField.build_from(
                _("String"),
                "string",
                str,
                description=_("The output string."),
            )
        ],
    )

    def __init__(self, separator: str = "\n", **kwargs):
        """Create a new ChunksToStringOperator."""
        self._separator = separator
        super().__init__(**kwargs)

    async def map(self, chunks: List[Chunk]) -> str:
        """Map the chunks to string."""
        return self._separator.join([chunk.content for chunk in chunks])



from typing import List
from unittest.mock import MagicMock, patch

import pytest

import dbgpt_ext
from dbgpt.core import Chunk
from dbgpt_ext.rag.retriever.db_schema import DBSchemaRetriever


@pytest.fixture
def mock_db_connection():
    return MagicMock()


@pytest.fixture
def mock_table_vector_store_connector():
    mock_connector = MagicMock()
    mock_connector.vector_store_config.name = "table_name"
    mock_connector.similar_search_with_scores.return_value = [
        Chunk(content="Table summary")
    ] * 4
    return mock_connector


@pytest.fixture
def mock_field_vector_store_connector():
    mock_connector = MagicMock()
    mock_connector.similar_search_with_scores.return_value = [
        Chunk(content="Field summary")
    ] * 4
    return mock_connector


@pytest.fixture
def dbstruct_retriever(
    mock_db_connection,
    mock_table_vector_store_connector,
    mock_field_vector_store_connector,
):
    return DBSchemaRetriever(
        connector=mock_db_connection,
        table_vector_store_connector=mock_table_vector_store_connector,
        field_vector_store_connector=mock_field_vector_store_connector,
    )


def mock_parse_db_summary() -> str:
    """Patch _parse_db_summary method."""
    return "Table summary"


# Mocking the _parse_db_summary method in your test function
@patch.object(
    dbgpt_ext.rag.summary.rdbms_db_summary, "_parse_db_summary", mock_parse_db_summary
)
def test_retrieve_with_mocked_summary(dbstruct_retriever):
    query = "Table summary"
    chunks: List[Chunk] = dbstruct_retriever._retrieve(query)
    assert isinstance(chunks[0], Chunk)
    assert chunks[0].content == "Table summary"


async def async_mock_parse_db_summary() -> str:
    """Asynchronous patch for _parse_db_summary method."""
    return "Table summary"



"""Module of RAG storage."""

from typing import Tuple, Type


def _import_pgvector() -> Tuple[Type, Type]:
    from dbgpt_ext.storage.vector_store.pgvector_store import (
        PGVectorConfig,
        PGVectorStore,
    )

    return PGVectorStore, PGVectorConfig


def _import_milvus() -> Tuple[Type, Type]:
    from dbgpt_ext.storage.vector_store.milvus_store import (
        MilvusStore,
        MilvusVectorConfig,
    )

    return MilvusStore, MilvusVectorConfig


def _import_chroma() -> Tuple[Type, Type]:
    from dbgpt_ext.storage.vector_store.chroma_store import (
        ChromaStore,
        ChromaVectorConfig,
    )

    return ChromaStore, ChromaVectorConfig


def _import_weaviate() -> Tuple[Type, Type]:
    from dbgpt_ext.storage.vector_store.weaviate_store import (
        WeaviateStore,
        WeaviateVectorConfig,
    )

    return WeaviateStore, WeaviateVectorConfig


def _import_oceanbase() -> Tuple[Type, Type]:
    from dbgpt_ext.storage.vector_store.oceanbase_store import (
        OceanBaseConfig,
        OceanBaseStore,
    )

    return OceanBaseStore, OceanBaseConfig


def _import_elastic() -> Tuple[Type, Type]:
    from dbgpt_ext.storage.vector_store.elastic_store import (
        ElasticsearchStoreConfig,
        ElasticStore,
    )

    return ElasticStore, ElasticsearchStoreConfig


def _import_builtin_knowledge_graph() -> Tuple[Type, Type]:
    from dbgpt_ext.storage.knowledge_graph.knowledge_graph import (
        BuiltinKnowledgeGraph,
        BuiltinKnowledgeGraphConfig,
    )

    return BuiltinKnowledgeGraph, BuiltinKnowledgeGraphConfig


def _import_community_summary_knowledge_graph() -> Tuple[Type, Type]:
    from dbgpt_ext.storage.knowledge_graph.community_summary import (
        CommunitySummaryKnowledgeGraph,
        CommunitySummaryKnowledgeGraphConfig,
    )

    return CommunitySummaryKnowledgeGraph, CommunitySummaryKnowledgeGraphConfig


def _import_openspg() -> Tuple[Type, Type]:
    from dbgpt_ext.storage.knowledge_graph.open_spg import OpenSPG, OpenSPGConfig

    return OpenSPG, OpenSPGConfig


def _import_full_text() -> Tuple[Type, Type]:
    from dbgpt_ext.storage.full_text.elasticsearch import ElasticDocumentStore
    from dbgpt_ext.storage.vector_store.elastic_store import ElasticsearchStoreConfig

    return ElasticDocumentStore, ElasticsearchStoreConfig


def _select_rag_storage(name: str) -> Tuple[Type, Type]:
    if name == "Chroma":
        return _import_chroma()
    elif name == "Milvus":
        return _import_milvus()
    elif name == "Weaviate":
        return _import_weaviate()
    elif name == "PGVector":
        return _import_pgvector()
    elif name == "OceanBase":
        return _import_oceanbase()
    elif name == "ElasticSearch":
        return _import_elastic()
    elif name == "KnowledgeGraph":
        return _import_builtin_knowledge_graph()
    elif name == "CommunitySummaryKnowledgeGraph":
        return _import_community_summary_knowledge_graph()
    elif name == "OpenSPG":
        return _import_openspg()
    elif name == "FullText":
        return _import_full_text()
    else:
        raise AttributeError(f"Could not find: {name}")


_HAS_SCAN = False


def scan_storage_configs():
    """Scan storage configs."""
    from dbgpt.storage.base import IndexStoreConfig
    from dbgpt.util.module_utils import ModelScanner, ScannerConfig

    global _HAS_SCAN

    if _HAS_SCAN:
        return
    modules = [
        "dbgpt_ext.storage.vector_store",
        "dbgpt_ext.storage.knowledge_graph",
        "dbgpt_ext.storage.graph_store",
    ]

    scanner = ModelScanner[IndexStoreConfig]()
    for module in modules:
        config = ScannerConfig(
            module_path=module,
            base_class=IndexStoreConfig,
        )
        scanner.scan_and_register(config)
    _HAS_SCAN = True
    return scanner.get_registered_items()


__vector_store__ = [
    "Chroma",
    "Milvus",
    "Weaviate",
    "OceanBase",
    "PGVector",
    "ElasticSearch",
]

__knowledge_graph__ = ["KnowledgeGraph", "CommunitySummaryKnowledgeGraph", "OpenSPG"]

__document_store__ = ["FullText"]

__all__ = __vector_store__ + __knowledge_graph__ + __document_store__



"""Oracle connector using python-oracledb."""

from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple, Type
from urllib.parse import quote_plus

from sqlalchemy import text

from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.datasource.rdbms.base import RDBMSConnector, RDBMSDatasourceParameters
from dbgpt.util.i18n_utils import _


@auto_register_resource(
    label=_("Oracle datasource"),
    category=ResourceCategory.DATABASE,
    tags={"order": TAGS_ORDER_HIGH},
    description=_(
        "Enterprise-grade relational database with oracledb driver (python-oracledb)."
    ),
)
@dataclass
class OracleParameters(RDBMSDatasourceParameters):
    """Oracle connection parameters."""

    __type__ = "oracle"

    driver: str = field(
        default="oracle+oracledb",  # ✅ 使用 python-oracledb 驱动
        metadata={
            "help": _("Driver name for Oracle, default is oracle+oracledb."),
        },
    )

    service_name: Optional[str] = field(
        default=None,
        metadata={
            "help": _("Oracle service name (alternative to SID)."),
        },
    )

    sid: Optional[str] = field(
        default=None,
        metadata={
            "help": _("Oracle SID (System ID, alternative to service name)."),
        },
    )

    def db_url(self, ssl: bool = False, charset: Optional[str] = None) -> str:
        if self.service_name:
            dsn = (
                f"(DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST={self.host})"
                f"(PORT={self.port}))(CONNECT_DATA=(SERVICE_NAME={self.service_name})))"
            )
        elif self.sid:
            dsn = (
                f"(DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST={self.host})"
                f"(PORT={self.port}))(CONNECT_DATA=(SID={self.sid})))"
            )
        else:
            raise ValueError("Either service_name or sid must be provided for Oracle.")

        return f"{self.driver}://{self.user}:{self.password}@{dsn}"

    def create_connector(self) -> "OracleConnector":
        return OracleConnector.from_parameters(self)


class OracleConnector(RDBMSConnector):
    db_type: str = "oracle"
    db_dialect: str = "oracle"
    driver: str = "oracle+oracledb"

    @classmethod
    def param_class(cls) -> Type[RDBMSDatasourceParameters]:
        return OracleParameters

    @classmethod
    def from_uri_db(
        cls,
        host: str,
        port: int,
        user: str,
        pwd: str,
        sid: Optional[str] = None,
        service_name: Optional[str] = None,
        engine_args: Optional[dict] = None,
        **kwargs,
    ) -> "OracleConnector":
        if not sid and not service_name:
            raise ValueError("Must provide either sid or service_name")

        if service_name:
            dsn = (
                f"(DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)"
                f"(HOST={host})(PORT={port}))(CONNECT_DATA=(SERVICE_NAME={service_name})))"
            )
        else:
            dsn = (
                f"(DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST={host})"
                f"(PORT={port}))(CONNECT_DATA=(SID={sid})))"
            )

        bm_pwd = quote_plus(pwd)
        db_url = f"{cls.driver}://{user}:{bm_pwd}@{dsn}"

        return cls.from_uri(db_url, engine_args=engine_args, **kwargs)

    def get_simple_fields(self, table_name):
        """Get column fields about specified table."""
        return self.get_fields(table_name)

    def get_fields(self, table_name: str, db_name=None) -> List[Tuple]:
        with self.session_scope() as session:
            query = f"""
                SELECT col.column_name,
                       col.data_type,
                       col.data_default,
                       col.nullable,
                       comm.comments
                FROM user_tab_columns col
                LEFT JOIN user_col_comments comm
                ON col.table_name = comm.table_name
                AND col.column_name = comm.column_name
                WHERE col.table_name = '{table_name.upper()}'
            """
            result = session.execute(text(query))
            return result.fetchall()

    def get_charset(self) -> str:
        with self.session_scope() as session:
            cursor = session.execute(
                text(
                    "SELECT VALUE FROM NLS_DATABASE_PARAMETERS "
                    "WHERE PARAMETER = 'NLS_CHARACTERSET'"
                )
            )
            return cursor.fetchone()[0]

    def get_grants(self):
        with self.session_scope() as session:
            cursor = session.execute(text("SELECT privilege FROM user_sys_privs"))
            return cursor.fetchall()

    def get_users(self) -> List[Tuple[str, None]]:
        with self.session_scope() as session:
            cursor = session.execute(text("SELECT username FROM all_users"))
            return [(row[0], None) for row in cursor.fetchall()]

    def get_database_names(self) -> List[str]:
        with self.session_scope() as session:
            is_cdb = session.execute(text("SELECT CDB FROM V$DATABASE")).fetchone()[0]
            if is_cdb == "YES":
                pdbs = session.execute(
                    text("SELECT NAME FROM V$PDBS WHERE OPEN_MODE = 'READ WRITE'")
                ).fetchall()
                return [name[0] for name in pdbs]
            else:
                return [
                    session.execute(
                        text("SELECT sys_context('USERENV', 'CON_NAME') FROM dual")
                    ).fetchone()[0]
                ]

    def get_table_comments(self, db_name: str) -> List[Tuple[str, str]]:
        with self.session_scope() as session:
            result = session.execute(
                text("SELECT table_name, comments FROM user_tab_comments")
            )
            return [(row[0], row[1]) for row in result.fetchall()]

    def get_table_comment(self, table_name: str) -> Dict:
        with self.session_scope() as session:
            cursor = session.execute(
                text(
                    f"SELECT comments FROM user_tab_comments "
                    f"WHERE table_name = '{table_name.upper()}'"
                )
            )
            row = cursor.fetchone()
            return {"text": row[0] if row else ""}

    def get_column_comments(
        self, db_name: str, table_name: str
    ) -> List[Tuple[str, str]]:
        with self.session_scope() as session:
            cursor = session.execute(
                text(f"""
                    SELECT column_name, comments
                    FROM user_col_comments
                    WHERE table_name = '{table_name.upper()}'
                """)
            )
            return [(row[0], row[1]) for row in cursor.fetchall()]

    def get_collation(self) -> str:
        with self.session_scope() as session:
            cursor = session.execute(
                text(
                    "SELECT value FROM NLS_DATABASE_PARAMETERS "
                    "WHERE parameter = 'NLS_SORT'"
                )
            )
            return cursor.fetchone()[0]



"""Hive Connector."""

from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Type, cast
from urllib.parse import quote
from urllib.parse import quote_plus as urlquote

from sqlalchemy import create_engine

from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.datasource.parameter import BaseDatasourceParameters
from dbgpt.datasource.rdbms.base import RDBMSConnector
from dbgpt.util.i18n_utils import _


@auto_register_resource(
    label=_("Apache Hive datasource"),
    category=ResourceCategory.DATABASE,
    tags={"order": TAGS_ORDER_HIGH},
    description=_("A distributed fault-tolerant data warehouse system."),
)
@dataclass
class HiveParameters(BaseDatasourceParameters):
    """Hive connection parameters."""

    __type__ = "hive"

    # Basic connection parameters
    host: str = field(metadata={"help": _("Hive server host")})
    port: int = field(
        default=10000, metadata={"help": _("Hive server port, default 10000")}
    )
    database: str = field(
        default="default", metadata={"help": _("Database name, default 'default'")}
    )

    # Authentication parameters
    auth: str = field(
        default="NONE",
        metadata={
            "help": _("Authentication mode: NONE, NOSASL, LDAP, KERBEROS, CUSTOM"),
            "valid_values": ["NONE", "NOSASL", "LDAP", "KERBEROS", "CUSTOM"],
        },
    )
    username: str = field(
        default="", metadata={"help": _("Username for authentication")}
    )
    password: str = field(
        default="",
        metadata={
            "help": _("Password for LDAP or CUSTOM auth"),
            "tags": "privacy",
        },
    )

    # Kerberos parameters
    kerberos_service_name: str = field(
        default="hive", metadata={"help": _("Kerberos service name")}
    )

    # Transport parameters
    transport_mode: str = field(
        default="binary", metadata={"help": _("Transport mode: binary or http")}
    )
    # http_path: str = field(
    #     default="", metadata={"help": _("HTTP path for HTTP transport mode")}
    # )
    driver: str = field(
        default="hive",
        metadata={
            "help": _("Driver name for Hive, default is hive."),
        },
    )

    def engine_args(self) -> Optional[Dict[str, Any]]:
        """Get engine args."""
        connect_args = {"auth": self.auth}
        # username and password are not required for NONE and NOSASL
        if self.username:
            connect_args["username"] = self.username
        if self.password and self.auth in ("LDAP", "CUSTOM"):
            connect_args["password"] = self.password
        if self.auth == "KERBEROS":
            connect_args["kerberos_service_name"] = self.kerberos_service_name
        return {
            "connect_args": {k: v for k, v in connect_args.items() if v},
        }

    def create_connector(self) -> "HiveConnector":
        """Create Hive connector."""
        return HiveConnector.from_parameters(self)

    def db_url(self, ssl: bool = False, charset: Optional[str] = None):
        """Return database engine url."""
        if self.driver:
            scheme = self.driver
        elif self.transport_mode == "http":
            scheme = "hive+http"
        else:
            scheme = "hive"

        if self.username and self.password:
            auth_str = f"{quote(self.username)}:{urlquote(self.password)}@"
        else:
            auth_str = ""
        return f"{scheme}://{auth_str}{self.host}:{str(self.port)}/{self.database}"


class HiveConnector(RDBMSConnector):
    """Hive connector."""

    db_type: str = "hive"
    """db driver"""
    driver: str = "hive"
    """db dialect"""
    dialect: str = "hive"

    @classmethod
    def param_class(cls) -> Type[HiveParameters]:
        """Return the parameter class."""
        return HiveParameters

    @classmethod
    def from_parameters(cls, parameters: HiveParameters) -> "HiveConnector":
        """Create connector from parameters.

        More details:
        https://github.com/apache/kyuubi/blob/master/python/pyhive/hive.py
        """
        db_url = parameters.db_url()
        engine_args = parameters.engine_args() or {}
        return cls(create_engine(db_url, **engine_args))

    @classmethod
    def from_uri_db(
        cls,
        host: str,
        port: int,
        user: str,
        pwd: str,
        db_name: str,
        engine_args: Optional[dict] = None,
        **kwargs: Any,
    ) -> "HiveConnector":
        """Create a new HiveConnector from host, port, user, pwd, db_name."""
        db_url: str = f"{cls.driver}://{host}:{str(port)}/{db_name}"
        if user and pwd:
            db_url = (
                f"{cls.driver}://{quote(user)}:{urlquote(pwd)}@{host}:{str(port)}/"
                f"{db_name}"
            )
        return cast(HiveConnector, cls.from_uri(db_url, engine_args, **kwargs))

    def table_simple_info(self):
        """Get table simple info."""
        return []

    def get_users(self):
        """Get users."""
        return []

    def get_grants(self):
        """Get grants."""
        return []

    def get_collation(self):
        """Get collation."""
        return "UTF-8"

    def get_charset(self):
        """Get character_set of current database."""
        return "UTF-8"

    def _format_sql(self, sql: str) -> str:
        """Format sql."""
        sql = super()._format_sql(sql)
        # remove ';' at the end of sql
        return sql.rstrip(";")



"""DBSchemaAssembler."""

from typing import Any, List, Optional

from dbgpt.core import Chunk, Embeddings
from dbgpt.datasource.base import BaseConnector
from dbgpt.rag.embedding.embedding_factory import DefaultEmbeddingFactory
from dbgpt.storage.vector_store.base import VectorStoreBase

from ..assembler.base import BaseAssembler
from ..chunk_manager import ChunkParameters
from ..knowledge.datasource import DatasourceKnowledge
from ..retriever.db_schema import DBSchemaRetriever


class DBSchemaAssembler(BaseAssembler):
    """DBSchemaAssembler.

    Example:
        .. code-block:: python

            from dbgpt_ext.datasource.rdbms.conn_sqlite import SQLiteTempConnector
            from dbgpt_ext.rag.assembler.db_schema import DBSchemaAssembler
            from dbgpt.storage.vector_store.base import VectorStoreBase
            from dbgpt_ext.storage.vector_store.chroma_store import ChromaVectorConfig

            connection = SQLiteTempConnector.create_temporary_db()
            assembler = DBSchemaAssembler.load_from_connection(
                connector=connection,
                embedding_model=embedding_model_path,
            )
            assembler.persist()
            # get db struct retriever
            retriever = assembler.as_retriever(top_k=3)
    """

    def __init__(
        self,
        connector: BaseConnector,
        table_vector_store_connector: VectorStoreBase,
        field_vector_store_connector: Optional[VectorStoreBase] = None,
        chunk_parameters: Optional[ChunkParameters] = None,
        embedding_model: Optional[str] = None,
        embeddings: Optional[Embeddings] = None,
        max_seq_length: int = 512,
        **kwargs: Any,
    ) -> None:
        """Initialize with Embedding Assembler arguments.

        Args:
            connector: (BaseConnector) BaseConnector connection.
            table_vector_store_connector: VectorStoreConnector to load
                                        and retrieve table info.
            field_vector_store_connector: VectorStoreConnector to load
                                        and retrieve field info.
            chunk_manager: (Optional[ChunkManager]) ChunkManager to use for chunking.
            embedding_model: (Optional[str]) Embedding model to use.
            embeddings: (Optional[Embeddings]) Embeddings to use.
        """
        self._connector = connector
        self._table_vector_store_connector = table_vector_store_connector
        self._field_vector_store_connector = field_vector_store_connector
        self._embedding_model = embedding_model
        if self._embedding_model and not embeddings:
            embeddings = DefaultEmbeddingFactory(
                default_model_name=self._embedding_model
            ).create(self._embedding_model)

        knowledge = DatasourceKnowledge(connector, model_dimension=max_seq_length)
        super().__init__(
            knowledge=knowledge,
            chunk_parameters=chunk_parameters,
            **kwargs,
        )

    @classmethod
    def load_from_connection(
        cls,
        connector: BaseConnector,
        table_vector_store_connector: VectorStoreBase,
        field_vector_store_connector: Optional[VectorStoreBase] = None,
        chunk_parameters: Optional[ChunkParameters] = None,
        embedding_model: Optional[str] = None,
        embeddings: Optional[Embeddings] = None,
        max_seq_length: int = 512,
    ) -> "DBSchemaAssembler":
        """Load document embedding into vector store from path.

        Args:
            connector: (BaseConnector) BaseConnector connection.
            table_vector_store_connector: used to load table chunks.
            field_vector_store_connector: used to load field chunks
                                        if field in table is too much.
            chunk_parameters: (Optional[ChunkParameters]) ChunkManager to use for
                chunking.
            embedding_model: (Optional[str]) Embedding model to use.
            embeddings: (Optional[Embeddings]) Embeddings to use.
            max_seq_length: Embedding model max sequence length
        Returns:
             DBSchemaAssembler
        """
        return cls(
            connector=connector,
            table_vector_store_connector=table_vector_store_connector,
            field_vector_store_connector=field_vector_store_connector,
            embedding_model=embedding_model,
            chunk_parameters=chunk_parameters,
            embeddings=embeddings,
            max_seq_length=max_seq_length,
        )

    def get_chunks(self) -> List[Chunk]:
        """Return chunk ids."""
        return self._chunks

    def persist(self, **kwargs: Any) -> List[str]:
        """Persist chunks into vector store.

        Returns:
            List[str]: List of chunk ids.
        """
        table_chunks, field_chunks = [], []
        for chunk in self._chunks:
            metadata = chunk.metadata
            if metadata.get("separated"):
                if metadata.get("part") == "table":
                    table_chunks.append(chunk)
                else:
                    field_chunks.append(chunk)
            else:
                table_chunks.append(chunk)

        if self._field_vector_store_connector and field_chunks:
            self._field_vector_store_connector.load_document_with_limit(field_chunks)
        return self._table_vector_store_connector.load_document_with_limit(table_chunks)

    def _extract_info(self, chunks) -> List[Chunk]:
        """Extract info from chunks."""
        return []

    def as_retriever(self, top_k: int = 4, **kwargs) -> DBSchemaRetriever:
        """Create DBSchemaRetriever.

        Args:
            top_k(int): default 4.

        Returns:
            DBSchemaRetriever
        """
        return DBSchemaRetriever(
            top_k=top_k,
            connector=self._connector,
            is_embeddings=True,
            table_vector_store_connector=self._table_vector_store_connector,
            field_vector_store_connector=self._field_vector_store_connector,
        )



"""Qianfan Embeddings module."""

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Type

from dbgpt._private.pydantic import BaseModel, ConfigDict, Field
from dbgpt.core import EmbeddingModelMetadata, Embeddings
from dbgpt.core.interface.parameter import EmbeddingDeployModelParameters
from dbgpt.model.adapter.base import register_embedding_adapter
from dbgpt.util.i18n_utils import _


@dataclass
class QianfanEmbeddingDeployModelParameters(EmbeddingDeployModelParameters):
    """Qianfan Embeddings deploy model parameters."""

    provider: str = "proxy/qianfan"

    api_key: Optional[str] = field(
        default=None,
        metadata={"help": _("The API key for the embeddings API.")},
    )
    api_secret: Optional[str] = field(
        default=None,
        metadata={
            "help": _("The Secret key for the embeddings API. It's the sk for qianfan.")
        },
    )
    backend: Optional[str] = field(
        default=None,
        metadata={
            "help": _(
                "The real model name to pass to the provider, default is None. If "
                "backend is None, use name as the real model name."
            ),
        },
    )

    @property
    def real_provider_model_name(self) -> str:
        """Get the real provider model name."""
        return self.backend or self.name


class QianFanEmbeddings(BaseModel, Embeddings):
    """Baidu Qianfan Embeddings embedding models.

    Embed:
       .. code-block:: python

           # embed the documents
           vectors = embeddings.embed_documents([text1, text2, ...])

           # embed the query
           vectors = embeddings.embed_query(text)

    """  # noqa: E501

    client: Any
    chunk_size: int = 16
    endpoint: str = ""
    """Endpoint of the Qianfan Embedding, required if custom model used."""
    model_config = ConfigDict(arbitrary_types_allowed=True, protected_namespaces=())
    api_key: Optional[str] = Field(
        default=None, description="The API key for the embeddings API."
    )
    api_secret: Optional[str] = Field(
        default=None, description="The Secret key for the embeddings API."
    )
    """Model name
    you could get from https://cloud.baidu.com/doc/WENXINWORKSHOP/s/Nlks5zkzu

    for now, we support Embedding-V1 and
    - Embedding-V1 （默认模型）
    - bge-large-en
    - bge-large-zh

    preset models are mapping to an endpoint.
    `model` will be ignored if `endpoint` is set
    """
    model_name: str = Field(
        default="text-embedding-v1", description="The name of the model to use."
    )
    init_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """init kwargs for qianfan client init, such as `query_per_second` which is
        associated with qianfan resource object to limit QPS"""

    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """extra params for model invoke using with `do`."""

    def __init__(self, **kwargs):
        """Initialize the QianFanEmbeddings."""
        try:
            import qianfan
        except ImportError as exc:
            raise ValueError(
                "Could not import python package: qianfan. "
                "Please install qianfan by running `pip install qianfan`."
            ) from exc

        qianfan_ak = kwargs.get("api_key")
        qianfan_sk = kwargs.get("api_secret")
        model_name = kwargs.get("model_name")

        if not qianfan_ak or not qianfan_sk or not model_name:
            raise ValueError(
                "API key, API secret, and model name are required to initialize "
                "QianFanEmbeddings."
            )

        params = {
            "model": model_name,
            "ak": qianfan_ak,
            "sk": qianfan_sk,
        }

        # Initialize the qianfan.Embedding client
        kwargs["client"] = qianfan.Embedding(**params)
        super().__init__(**kwargs)

    @classmethod
    def param_class(cls) -> Type[QianfanEmbeddingDeployModelParameters]:
        """Get the parameter class."""
        return QianfanEmbeddingDeployModelParameters

    @classmethod
    def from_parameters(
        cls, parameters: QianfanEmbeddingDeployModelParameters
    ) -> "QianFanEmbeddings":
        """Create an embedding model from parameters."""
        return cls(
            api_key=parameters.api_key,
            api_secret=parameters.api_secret,
            model_name=parameters.real_provider_model_name,
        )

    def embed_query(self, text: str) -> List[float]:
        """Compute query embeddings using a QianFan embedding model."""
        resp = self.embed_documents([text])
        return resp[0]

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """
        Embeds a list of text documents using the AutoVOT algorithm.

        Args:
            texts (List[str]): A list of text documents to embed.

        Returns:
            List[List[float]]: A list of embeddings for each document in the input list.
                            Each embedding is represented as a list of float values.
        """
        text_in_chunks = [
            texts[i : i + self.chunk_size]
            for i in range(0, len(texts), self.chunk_size)
        ]
        lst = []
        for chunk in text_in_chunks:
            resp = self.client.do(texts=chunk, **self.model_kwargs)
            lst.extend([res["embedding"] for res in resp["data"]])
        return lst


register_embedding_adapter(
    QianFanEmbeddings,
    supported_models=[
        EmbeddingModelMetadata(
            model="embedding-v1",
            description=_("Embedding-V1 by Baidu Qianfan. "),
            link="https://cloud.baidu.com/doc/WENXINWORKSHOP/s/alj562vvu",
        )
    ],
)



"""Knowledge Factory to create knowledge from file path and url."""

from typing import Dict, List, Optional, Type, Union

from dbgpt.rag.knowledge.base import Knowledge, KnowledgeType
from dbgpt_ext.rag.knowledge.string import StringKnowledge
from dbgpt_ext.rag.knowledge.url import URLKnowledge


class KnowledgeFactory:
    """Knowledge Factory to create knowledge from file path and url."""

    def __init__(
        self,
        file_path: Optional[str] = None,
        knowledge_type: Optional[KnowledgeType] = KnowledgeType.DOCUMENT,
        metadata: Optional[Dict[str, Union[str, List[str]]]] = None,
    ):
        """Create Knowledge Factory with file path and knowledge type.

        Args:
            file_path(str, optional): file path
            knowledge_type(KnowledgeType, optional): knowledge type
        """
        self._file_path = file_path
        self._knowledge_type = knowledge_type
        self._metadata = metadata

    @classmethod
    def create(
        cls,
        datasource: str = "",
        knowledge_type: KnowledgeType = KnowledgeType.DOCUMENT,
        metadata: Optional[Dict[str, Union[str, List[str]]]] = None,
    ):
        """Create knowledge from file path, url or text.

        Args:
             datasource: path of the file to convert
             knowledge_type: type of knowledge
             metadata: Optional[Dict[str, Union[str, List[str]]]]

        Examples:
            .. code-block:: python

                from dbgpt.rag.knowledge.factory import KnowledgeFactory

                url_knowlege = KnowledgeFactory.create(
                    datasource="https://www.baidu.com", knowledge_type=KnowledgeType.URL
                )
                doc_knowlege = KnowledgeFactory.create(
                    datasource="path/to/document.pdf",
                    knowledge_type=KnowledgeType.DOCUMENT,
                )

        """
        match knowledge_type:
            case KnowledgeType.DOCUMENT:
                return cls.from_file_path(
                    file_path=datasource,
                    knowledge_type=knowledge_type,
                    metadata=metadata,
                )
            case KnowledgeType.URL:
                return cls.from_url(url=datasource, knowledge_type=knowledge_type)
            case KnowledgeType.TEXT:
                return cls.from_text(
                    text=datasource, knowledge_type=knowledge_type, metadata=metadata
                )
            case _:
                raise Exception(f"Unsupported knowledge type '{knowledge_type}'")

    @classmethod
    def from_file_path(
        cls,
        file_path: str = "",
        knowledge_type: Optional[KnowledgeType] = KnowledgeType.DOCUMENT,
        metadata: Optional[Dict[str, Union[str, List[str]]]] = None,
    ) -> Knowledge:
        """Create knowledge from path.

        Args:
            param file_path: path of the file to convert
            param knowledge_type: type of knowledge

        Examples:
            .. code-block:: python

                from dbgpt.rag.knowledge.factory import KnowledgeFactory

                doc_knowlege = KnowledgeFactory.create(
                    datasource="path/to/document.pdf",
                    knowledge_type=KnowledgeType.DOCUMENT,
                )

        """
        factory = cls(file_path=file_path, knowledge_type=knowledge_type)
        return factory._select_document_knowledge(
            file_path=file_path, knowledge_type=knowledge_type, metadata=metadata
        )

    @staticmethod
    def from_url(
        url: str = "",
        knowledge_type: KnowledgeType = KnowledgeType.URL,
    ) -> Knowledge:
        """Create knowledge from url.

        Args:
            param url: url of the file to convert
            param knowledge_type: type of knowledge

        Examples:
            .. code-block:: python

                from dbgpt.rag.knowledge.factory import KnowledgeFactory

                url_knowlege = KnowledgeFactory.create(
                    datasource="https://www.baidu.com", knowledge_type=KnowledgeType.URL
                )
        """
        return URLKnowledge(
            url=url,
            knowledge_type=knowledge_type,
        )

    @staticmethod
    def from_text(
        text: str = "",
        knowledge_type: KnowledgeType = KnowledgeType.TEXT,
        metadata: Optional[Dict[str, Union[str, List[str]]]] = None,
    ) -> Knowledge:
        """Create knowledge from text.

        Args:
            param text: text to convert
            param knowledge_type: type of knowledge
        """
        return StringKnowledge(
            text=text,
            knowledge_type=knowledge_type,
            metadata=metadata,
        )

    def _select_document_knowledge(self, **kwargs):
        """Select document knowledge from file path."""
        extension = self._file_path.rsplit(".", 1)[-1]
        knowledge_classes = self._get_knowledge_subclasses()
        implementation = None
        for cls in knowledge_classes:
            if cls.document_type() and cls.document_type().value == extension:
                implementation = cls(**kwargs)
        if implementation is None:
            raise Exception(f"Unsupported knowledge document type '{extension}'")
        return implementation

    @classmethod
    def all_types(cls):
        """Get all knowledge types."""
        return [knowledge.type().value for knowledge in cls._get_knowledge_subclasses()]

    @classmethod
    def subclasses(cls) -> List["Type[Knowledge]"]:
        """Get all knowledge subclasses."""
        return cls._get_knowledge_subclasses()

    @staticmethod
    def _get_knowledge_subclasses() -> List["Type[Knowledge]"]:
        """Get all knowledge subclasses."""
        from dbgpt.rag.knowledge.base import Knowledge  # noqa: F401

        from .csv import CSVKnowledge  # noqa: F401
        from .datasource import DatasourceKnowledge  # noqa: F401
        from .doc import Word97DocKnowledge  # noqa: F401
        from .docx import DocxKnowledge  # noqa: F401
        from .excel import ExcelKnowledge  # noqa: F401
        from .html import HTMLKnowledge  # noqa: F401
        from .markdown import MarkdownKnowledge  # noqa: F401
        from .pdf import PDFKnowledge  # noqa: F401
        from .pptx import PPTXKnowledge  # noqa: F401
        from .string import StringKnowledge  # noqa: F401
        from .txt import TXTKnowledge  # noqa: F401
        from .url import URLKnowledge  # noqa: F401

        return Knowledge.__subclasses__()



from typing import List
from unittest.mock import MagicMock, patch

import pytest

import dbgpt_ext
from dbgpt.core import Chunk
from dbgpt_ext.rag.retriever.db_schema import DBSchemaRetriever


@pytest.fixture
def mock_db_connection():
    return MagicMock()


@pytest.fixture
def mock_table_vector_store_connector():
    mock_connector = MagicMock()
    mock_connector.vector_store_config.name = "table_name"
    chunk = Chunk(
        content="table_name: user\ncomment: user about dbgpt",
        metadata={
            "field_num": 6,
            "part": "table",
            "separated": 1,
            "table_name": "user",
        },
    )
    mock_connector.similar_search_with_scores = MagicMock(return_value=[chunk])
    return mock_connector


@pytest.fixture
def mock_field_vector_store_connector():
    mock_connector = MagicMock()
    chunk1 = Chunk(
        content="name,age",
        metadata={
            "field_num": 6,
            "part": "field",
            "part_index": 0,
            "separated": 1,
            "table_name": "user",
        },
    )
    chunk2 = Chunk(
        content="address,gender",
        metadata={
            "field_num": 6,
            "part": "field",
            "part_index": 1,
            "separated": 1,
            "table_name": "user",
        },
    )
    chunk3 = Chunk(
        content="mail,phone",
        metadata={
            "field_num": 6,
            "part": "field",
            "part_index": 2,
            "separated": 1,
            "table_name": "user",
        },
    )
    mock_connector.similar_search_with_scores = MagicMock(
        return_value=[chunk1, chunk2, chunk3]
    )
    return mock_connector


@pytest.fixture
def dbstruct_retriever(
    mock_db_connection,
    mock_table_vector_store_connector,
    mock_field_vector_store_connector,
):
    return DBSchemaRetriever(
        connector=mock_db_connection,
        table_vector_store_connector=mock_table_vector_store_connector,
        field_vector_store_connector=mock_field_vector_store_connector,
        separator="--table-field-separator--",
    )


def mock_parse_db_summary() -> str:
    """Patch _parse_db_summary method."""
    return (
        "table_name: user\ncomment: user about dbgpt\n"
        "--table-field-separator--\n"
        "name,age\naddress,gender\nmail,phone"
    )


# Mocking the _parse_db_summary method in your test function
@patch.object(
    dbgpt_ext.rag.summary.rdbms_db_summary, "_parse_db_summary", mock_parse_db_summary
)
def test_retrieve_with_mocked_summary(dbstruct_retriever):
    query = "Table summary"
    chunks: List[Chunk] = dbstruct_retriever._retrieve(query)
    assert isinstance(chunks[0], Chunk)
    assert "-table-field-separator--" in chunks[0].content


def async_mock_parse_db_summary() -> str:
    """Asynchronous patch for _parse_db_summary method."""
    return (
        "table_name: user\ncomment: user about dbgpt\n"
        "--table-field-separator--\n"
        "name,age\naddress,gender\nmail,phone"
    )



"""Define the CommunityStore class."""

import asyncio
import logging
from typing import List, Optional

from dbgpt.storage.vector_store.base import VectorStoreBase
from dbgpt_ext.rag.transformer.community_summarizer import CommunitySummarizer
from dbgpt_ext.storage.knowledge_graph.community.base import (
    Community,
    GraphStoreAdapter,
)
from dbgpt_ext.storage.knowledge_graph.community.community_metastore import (
    BuiltinCommunityMetastore,
)

logger = logging.getLogger(__name__)


class CommunityStore:
    """CommunityStore Class."""

    def __init__(
        self,
        graph_store_adapter: GraphStoreAdapter,
        community_summarizer: CommunitySummarizer,
        vector_store: VectorStoreBase,
        index_name: Optional[str] = None,
        max_chunks_once_load: Optional[int] = 10,
        max_threads: Optional[int] = 1,
        top_k: Optional[int] = 5,
        score_threshold: Optional[float] = 0.7,
    ):
        """Initialize the CommunityStore class."""
        self._graph_store_adapter = graph_store_adapter
        self._community_summarizer = community_summarizer
        self._meta_store = BuiltinCommunityMetastore(
            vector_store=vector_store,
            index_name=index_name,
            max_chunks_once_load=max_chunks_once_load,
            max_threads=max_threads,
            top_k=top_k,
            score_threshold=score_threshold,
        )

    async def build_communities(self, batch_size: int = 1):
        """Discover communities."""
        community_ids = await self._graph_store_adapter.discover_communities()

        # summarize communities
        communities = []
        n_communities = len(community_ids)

        for i in range(0, n_communities, batch_size):
            batch_ids = community_ids[i : i + batch_size]
            batch_results = await asyncio.gather(
                *[self._summary_community(cid) for cid in batch_ids]
            )
            # filter out None returns
            communities.extend([c for c in batch_results if c is not None])

        # truncate then save new summaries
        await self._meta_store.truncate()
        await self._meta_store.save(communities)

    async def _summary_community(self, community_id: str) -> Optional[Community]:
        """Summarize single community."""
        community = await self._graph_store_adapter.get_community(community_id)
        if community is None or community.data is None:
            logger.warning(f"Community {community_id} is empty")
            return None

        graph = community.data.format()
        community.summary = (
            await self._community_summarizer.summarize(graph=graph) or ""
        )
        logger.info(f"Summarize community {community_id}: {community.summary[:50]}...")
        return community

    async def search_communities(self, query: str) -> List[Community]:
        """Search communities."""
        return await self._meta_store.search(query)

    def truncate(self):
        """Truncate community store."""
        logger.info("Truncate community metastore")
        self._meta_store.truncate()

        logger.info("Truncate community summarizer")
        self._community_summarizer.truncate()

        logger.info("Truncate graph")
        self._graph_store_adapter.truncate()

    def drop(self):
        """Drop community store."""
        logger.info("Remove community metastore")
        self._meta_store.drop()

        logger.info("Remove community summarizer")
        self._community_summarizer.drop()

        logger.info("Remove graph")
        self._graph_store_adapter.drop()
